import os
import poppy
import numpy as np

from . import utils
from .webbpsf_core import _log


#######  Custom Optics used in JWInstrument classes  #####


class NIRSpec_three_MSA_shutters(poppy.AnalyticOpticalElement):
    """ Three NIRSpec MSA shutters, adjacent vertically."""

    def getPhasor(self,wave):
        """ Compute the transmission inside/outside of the field stop.

        The area of an open shutter is 0.2 x 0.45, while the shutter pitch is 0.26x0.51
        The walls separating adjacent shutters are 0.06 arcsec wide.
        """

        msa_width = 0.2
        msa_height = 0.45
        msa_wall = 0.06

        if not isinstance(wave, poppy.Wavefront):
            raise ValueError("FieldStop getPhasor must be called with a Wavefront to define the spacing")
        assert (wave.planetype == poppy.poppy_core._IMAGE)

        y, x= wave.coordinates()
        #xnew =  x*np.cos(np.deg2rad(self.angle)) + y*np.sin(np.deg2rad(self.angle))
        #ynew = -x*np.sin(np.deg2rad(self.angle)) + y*np.cos(np.deg2rad(self.angle))
        #x,y = xnew, ynew


        self.transmission = np.zeros(wave.shape)
        # get the innermost shutter than spans the Y axis
        w_inside_1 = np.where( (abs(y) < (msa_height/2))  & (abs(x) < (msa_width/2)))
        self.transmission[w_inside_1] = 1
        # get the adjacent shutters one above and one below.
        w_inside_2 = np.where( (abs(y) > (msa_height/2)+msa_wall) & (abs(y) < msa_height*1.5+msa_wall)  & (abs(x) < (msa_width/2)))
        self.transmission[w_inside_2] = 1

        return self.transmission


class NIRSpec_MSA_open_grid(poppy.AnalyticOpticalElement):
    """ An infinite repeating region of the NIRSpec MSA grid"""

    def getPhasor(self,wave):
        """ Compute the transmission inside/outside of the field stop.

        The area of an open shutter is 0.2 x 0.45, while the shutter pitch is 0.26x0.51
        The walls separating adjacent shutters are 0.06 arcsec wide.
        """

        msa_width = 0.2
        msa_height = 0.45
        msa_wall = 0.06
        msa_x_pitch = 0.26
        msa_y_pitch = 0.51

        if not isinstance(wave, poppy.Wavefront):
            raise ValueError("FieldStop getPhasor must be called with a Wavefront to define the spacing")
        assert (wave.planetype == poppy.poppy_core._IMAGE)

        y, x= wave.coordinates()
        #xnew =  x*np.cos(np.deg2rad(self.angle)) + y*np.sin(np.deg2rad(self.angle))
        #ynew = -x*np.sin(np.deg2rad(self.angle)) + y*np.cos(np.deg2rad(self.angle))
        #x,y = xnew, ynew

        mask_vert_walls  = np.abs(np.mod(np.abs(x), msa_x_pitch) - (msa_x_pitch/2)) < msa_wall/2
        mask_horz_walls  = np.abs(np.mod(np.abs(y), msa_y_pitch) - (msa_y_pitch/2)) < msa_wall/2



        self.transmission = np.ones(wave.shape)
        self.transmission[mask_vert_walls] = 0
        self.transmission[mask_horz_walls] = 0

        return self.transmission


class NIRISS_GR700XD_Grism(poppy.AnalyticOpticalElement):
    """ Custom optic class to model the NIRISS SOSS grim GR700XD

    This includes both the pupil mask file and the cylindrical lens

    Based on inputs from Loic Albert, Anand Sivaramakrishnan, and Andre Martel
    In particular see FGS_TFI_UdM_035_RevD for details of the NIRISS GR700XD
    measurement, and JWST-STScI-003338 for detector orientation and layout.

    GRISM DESIGN:

    The grism (and cylinder) are per design rotated by 2 degrees so as to be able
    to sample an emission line across different pixel position along the spatial
    direction (kind of resampling the line and not be limited by intra pixel
    response).  

    From Loic Albert's NIRISS technical report:

        * surface sag for the cylinder: 3.994 micron peak
        * limited to 3.968 microns for the 26 mm FOV mask

    From Loic Albert's email to Marshall 2013-07-18:

            I do have an update concerning the geometry of the GR700XD pupil
            mask. It turns out that they clocked the grism by about 2.25 degrees wrt the
            OTE system of coordinates. However, the field mask did not follow and is still
            aligned along the OTE s.o.c. That was a mistake that fortunately does have much
            impact.

            Comdev is in the process of modelling a new mask for the
            Spare grism. Remember that we will swap the current FLight GR700XD for
            its Spare which offers much improved performances. The mask size will
            be a little different (rectangular) and this time will be clocked 2.25
            degrees along with the grism.

            The sign of the rotation of the grism will have to be
            devised by trying the 2 possibilities an looking at the resulting tilt
            of the monochromatic PSF and the position of that PSF on the detector.
            Attached is a simulation of what we expect based on my own PSF
            generator.

            The cylinder lens has a well characterized power (actually radius of curvature). The values are:
                current Flight: 22.85 meters
                Spare: 22.39 meters 

            Prism physical size: pupil is 26 mm on a side for the current prism, will be 28 mm for the spare

    From Loic Albert's email to Marshall 2013-09-19:

        The latest news on this front are:

        1 - The current Flight mask is attached. It is 26x26 mm. The mask and grism are
            *not* aligned along the same coordinates. That was a mistake. I'll forward you
            a message from Michael M., our optics expert at CSA. 

        2 - The Spare mask (likely the one which will fly) is not built yet. The mask
            will be aligned along the grism coordinate and both will be clocked 2.2 deg wrt
            the OTE.

        3 - A ghost analysis showed that the current grism clocking will suffer from
            large ghosts. So we are studying how to clock the Spare grism in its cell to
            minimize ghosts. Likely a 90 degrees rotation will be applied to baseline of
            point 2.

        From Michael.Maszkiewicz@asc-csa.gc.ca:

            As far as I understand now, we have two rotations in the as-built
            GR700. One rotation is for the prism-grism combo by 2 deg CCW, looking along
            the local +z axis, and the second rotation is for  the mask by 3.05 deg  but
            CW. As a result there is total 5.05 deg rotation between grism and its mask.
            See my annotations to your drawing attached.

    From Loic Albert's email to Marshall 2014-05-20:

        I should have pointed that the power assumed in my simulations for the cylindrical lens was off. It was one of the conclusions of CV1RR. The actual radius of curvature of the cylinder is 25.3 meters (rather than the smaller figure I used before).
   
     ORIENTATION:

        See Figure 2 of JWST-STScI-003338
        In "DMS" coordinates, as projected looking outwards onto the sky,
        The GR700XD grating trace is near the extreme right edge of the detector
        with long wavelengths closest to (2048,2048) and short wavelengths nearest (2048,0)
        (The raw detector coordinates are very different from this due to a 180 degree rotation)

        **PLEASE NOTE** that the DMS when processing spectral data performs an additional transformation:
            For spectral data, the science X-axis is aligned with the detector
            dispersion direction and the science frame Y-axis is at a right angle
            to the X-axis in a right-handed coordinate system (Swade 2003)

        We choose here to ignore that complication; WebbPSF simulates the 2D sky projected
        image in "Sci" coordinates in the terminology for SIAF from Lallo et al. 
        In this coordinate system, the dispersion from the cylinder lens is aligned 
        almost along V2 and the longer wavelengths are oriented toward +V3. 
        

   

    Parameters
    ----------
    which : string
        'initial' or 'spare'. Properties are hard coded. 
    """
    #
    #    transmission : string filename
    #        file for the pupil transmission function
    #    cylinder_sag_mm : float
    #        physical thickness of the cylindrical lens, in millimeters
    #    rotation_angle : float
    #        degrees clockwise for the orientation of the cylinder's dispersing axis. Default
    #        of 92.25 should be consistent with initial NIRISS girsm and spare, except for
    #        sign ambiguity.
    #    rotate_mask : bool
    #        should the field mask be rotated along with the cylinder? False for first gen initial 
    #        prism, true for expected spare replacement.

    def __init__(self, name='GR700XD', which='Bach',
            #cylinder_radius=22.85,  cylinder_sag_mm=4.0, rotation_angle=92.25, rotate_mask=False, transmission=None, 
            shift=None):
        # Initialize the base optical element with the pupil transmission and zero OPD


        if which=='LLNL':
            raise NotImplementedError("Rotated field mask for LLNL grism not yet implemented!")
        elif which=='Bach':
            transmission=os.path.join( utils.get_webbpsf_data_path(), "NIRISS/optics/MASKGR700XD.fits.gz")
        else:
            raise NotImplementedError("Unknown grating name:"+which)

        self.shift=shift
        poppy.AnalyticOpticalElement.__init__(self, name=name, planetype=poppy.poppy_core._PUPIL)

        # UPDATED NUMBERS 2013-07:
        # See Document FGS_TFI_UdM_035_RevD

        _log.debug("Computing properties for {0} grism".format(which))
        if which =='Bach':
            #---- Phase properties ---------------
            # 3.994 microns P-V over 27.02 mm measured (Loic's email)
            # This is **surface sag**, corresponding to P-V of 6.311 waves at lambda=632.8 nm.
            # should correspond to 3.698 microns over 26 mm clear aperture. 
            self.prism_size = 0.02702 # 27.02 millimeters for the physical prism
            self.prism_clear_aperture = 0.0260 # 26 mm clear aperture for the prism + mount
            self.cylinder_rotation_angle = 2 # was 2.25

            #self.cylinder_radius = 22.85 # radius of curvature  ; Nominal
            # but they discarded that and used 25.3 instead
            # From Lafreniere's wfe_cylindricallens.pro:
            #  "OVERRIDE PREVIOUS CASES AFTER CV1RR RESULTS:"
            self.cylinder_radius = 25.3 # radius of curvature

            #---- Amplitude Transmission / Pupil shape ---------------
            self.pupil_size_mm = 26.0
            # Note that the IDL code says 26 mm is 683.75 pixels using the assumed demagnification
            self.pupil_rotation_angle = 2.0

        else:
            # 5.8 microns P-V over 32.15 mm (Loic's email)
            # should correspond to 4.38 microns over 28 mm clear aperture
            self.cylinder_radius = 22.39 # radius of curvature
            self.prism_size = 0.03215 # millimeters for the physical prism
            self.prism_clear_aperture = 0.0280 # clear aperture for the prism + mount
            self.cylinder_rotation_angle = 2.25

        # We need to know the magnification scale of the NIRISS reimaged pupil
        # in order to compute the curvature in the full-pupil units that POPPY uses
        # internally

        # pupil magnification computed from 22 mm clear aperture reported = 
        # 857-169 pixels = 699 pixels in the 2D array which has scale =.00645604
        # = 4.44175 meters projected on the primary

        # 2014-05-21 but wait, that's actually 26 mm!
        # so the 699 pixels at 0.00645604 m/pixel = 4.512 meters implies the magnificationa 173 not 170
        # but, double wait, it's actually more like 687 pixels across rather than 699 so that makes it 170 again.

        # therefore the magnification is 0.1708 meters projected on the primary / mm in the NIRISS pupil
        #self.pupil_demagnification =  170.8367 # meters on the primary / meters in the NIRISS pupil
        #self.pupil_demagnification =  173.56 # meters on the primary / meters in the NIRISS pupil

        # Anand says: 
        #  nominally the circumscribing circle at the PW of NIRISS is ~40mm.  I use 39mm for the nrm, but it's slightly field-dependent.  Compare that to the 6.6... PM circle?
        self.pupil_demagnification = 6.6 / 0.040 # about 165

        # perform an initial population of the OPD array for display etc.
        tmp = self.getPhasor(poppy.Wavefront(2e-6))

    def makeCylinder(self, wave):
        """ Make an OPD array corresponding to the cylindrical weak lens
        used for defocusing the spectrum in the perpendicular-to-dispersion direction.
        """
 
        if isinstance(wave, poppy.Wavefront):
            wavelength=wave.wavelength
        else:
            wavelength=float(wave)
            wave =  poppy.Wavefront(wavelength=wave)

        # compute indices in pixels, relative to center of plane, with rotation
        # units of these are meters
        y, x = wave.coordinates()

        ang = np.deg2rad(self.cylinder_rotation_angle )
        x = np.cos(ang)*x - np.sin(ang)*y
        y = np.sin(ang)*x + np.cos(ang)*y

        _log.debug(" Rotating local grism axes by {0} degrees".format(self.cylinder_rotation_angle))

        # From IDL code by David Lafreniere:
        #  ;the cylindrical defocus
        #x=(dindgen(pupdim)-pupdim/2)#replicate(1,pupdim)
        #y0=(rpuppix^2+sag[s]^2)/(2*sag[s])
        #wfe1=y0-sqrt(y0^2-x^2)
        #if sag[s] lt 1.e-5 then wfe1=0.d0

        # Here I will just translate that to Python exactly, making use of the
        # variables here:

        # rpuppix = radius of pupil in pixels
        #rpuppix = self.amplitude_header['DIAM'] / self.amplitude_header['PUPLSCAL'] / 2
        # Calculate the radius of curvature of the cylinder, bsaed on 
        # the chord length and height 

        # In this case we're assuming the cylinder is precisely as wide as the projected
        # telescope pupil. This doesn't seem guaranteed:
        #  * actual chord length across cylinder: 27.02 mm. 
        #  * projected primary scale at NIRISS = ?

        _log.debug(" Computing GR700XD cylinder based on RoC: {0:.3g} meters".format(self.cylinder_radius))
        _log.debug(" Computing GR700XD cylinder based on pupil demagnification: {0:.3g} primary to grism".format(self.pupil_demagnification))


        # Compute the overall sag of the cylinder lens at its outer edge. This is not actually used, it's
        # just for cross-check of the values
        # the sag will depend on half the pupil size since that's the offset from center to edge
        sag0 = np.sqrt(self.cylinder_radius**2 - (self.prism_size/2)**2) - self.cylinder_radius
        _log.debug(" Computed GR700XD cylinder sag at lens outer edge (for cross check only): {0:.3g} meters".format(sag0))

        # now compute the spatially dependent sag of the cylinder, as projected onto the primary

        # what is the pupil scale at the *reimaged pupil* of the grism?
        pupil_scale_m_per_pix = 38.0255e-6 # Based on UdeM info in wfe_cylindricallens.pro
        #sag = np.sqrt(self.cylinder_radius**2 - (x*self.amplitude_header['PUPLSCAL']/self.pupil_demagnification)**2) - self.cylinder_radius
        sag = np.sqrt(self.cylinder_radius**2 - (x/self.pupil_demagnification)**2) - self.cylinder_radius
        #sag = self.cylinder_radius -  np.sqrt(self.cylinder_radius**2 - (x * pupil_scale_m_per_pix )**2 )


        # what we really want to do is take the physical properties of the as-built optic, and interpolate into that
        # to compute the OPD after remapping based on the pupil scale (and distortion?)
        #y0=(rpuppix**2+self.cylinder_sag**2)/(2*self.cylinder_sag)
        #wfe1=y0-np.sqrt(y0**2-x**2)

        _log.debug(" Cylinder P-V: {0:.4g} meters physical sag across full array".format(sag.max()-sag.min()) )

        #fits.writeto('py_opd.fits', sag*(self.ZnSe_index(wavelength) -1), clobber=True)
        # remove piston offset 
        #sag -= sag[wnz].min()   # normalize to 0 at the minimum
        #sag -= sag[wnz].mean()    # normalize around the mean
        sag[self.amplitude == 0] = 0 # no OPD in opaque regions (makes no difference in propagation but improves display)
        wnz = np.where(self.amplitude != 0) # this is just for display of the log messages:
        _log.debug(" Cylinder P-V: {0:.4g} meters physical sag across clear aperture".format(sag[wnz].max()-sag[wnz].min()) )


        # scale for index of refraction
        index = self.ZnS_index(wavelength)
        opd = sag *  ( index-1)
        _log.debug(" Scaling for ZnS index of refraction {0} at {1:.3g} microns".format(index, wavelength*1e6))
        _log.debug(" Cylinder P-V: {0:.4g} meters optical sag at {1:.3g} microns across clear aperture".format(opd[wnz].max()-opd[wnz].min(), wavelength*1e6) )
        return opd

    def makePupil(self, wave):
        """ Make array for the pupil obscuration appropriate to the grism
        """

        if isinstance(wave, poppy.Wavefront):
            wavelength=wave.wavelength
        else:
            wavelength=float(wave)
            wave =  poppy.Wavefront(wavelength=wave)
        y, x = wave.coordinates()
        ang = np.deg2rad(self.pupil_rotation_angle )
        x = np.cos(ang)*x - np.sin(ang)*y
        y = np.sin(ang)*x + np.cos(ang)*y

        _log.debug("Rotating local pupil mask axes by {0} degrees".format(self.cylinder_rotation_angle))
 
        pupil_halfsize_m = self.pupil_size_mm / 2  / 1000 * self.pupil_demagnification
        pupilmask = np.ones_like(x)
        pupilmask[np.abs(x) > pupil_halfsize_m] = 0
        pupilmask[np.abs(y) > pupil_halfsize_m] = 0

        return pupilmask


    def ZnS_index(self, wavelength, temperature=40):
        """ Return cryogenic index of refraction of ZnS (Cleartran)

        Based on IDL function index_cleartran provided by Loic Albert at U de Montreal
        Which was in turn based on Leviton and Fray 2013
        http://proceedings.spiedigitallibrary.org/proceeding.aspx?articleid=1744938
        doi:10.1117/12.2024817

        """
        lambda_micron = wavelength*1e6

        # Sellmeier dispersion model
        #From Leviton & Frey measurements (SPIE preprint) (assumes lambda in microns)
        S_1 = np.asarray([[3.35933,-5.12262e-4,1.01086e-5,-4.14798e-8,6.91051e-11]])
        S_2 = np.asarray([[0.706131,4.89603e-4,-8.91159e-6,3.81621e-8,-6.54805e-11]])
        S_3 = np.asarray([[4.02154,-2.93193e-2,2.31080e-4,-7.57289e-07,8.31188e-10]])
        S_ij = np.concatenate( (S_1,S_2,S_3), axis=0)
        lambda_1 = np.array([[0.161151, -8.93057E-06, 2.73286E-07, -1.23408E-09, 2.29917E-12]])
        lambda_2 = np.array([[0.282427, -4.66636E-05, 7.55906E-07, -2.77513E-09, 4.35237E-12]])
        lambda_3 = np.array([[41.1590, -0.161010, 1.23906E-03, -3.95895E-06, 4.16370E-09]])
        lambda_ij = np.concatenate((lambda_1,lambda_2,lambda_3))


        n2minus1 = 0.0
        T=temperature
        for i in range(3):
            S_i =       S_ij[i,0]      + S_ij[i,1]     *T + S_ij[i,2]     *T**2.0 + S_ij[i,3]     *T**3.0 + S_ij[i,4]     *T**4.0
            lambda_i =  lambda_ij[i,0] + lambda_ij[i,1]*T + lambda_ij[i,2]*T**2.0 + lambda_ij[i,3]*T**3.0 + lambda_ij[i,4]*T**4.0
            n2minus1 += S_i*lambda_micron**2.0/(lambda_micron**2.0 - lambda_i**2.0)

        cleartran_index = np.sqrt(1.0 + n2minus1)
        return cleartran_index


    def getPhasor(self, wave):
        """ Scale the cylindrical lens OPD appropriately for the current wavelength
            Then call the regular getphasor method of the parent class 

        """
        # Pupil shape part: 
        self.amplitude = self.makePupil(wave)

        # Phase part: 
        self.opd = self.makeCylinder(wave)

        # Return the OPD derived from them both
        return   poppy.OpticalElement.getPhasor(self, wave)


    def display(self, opd_vmax=6e-6, *args, **kwargs):
        "Same as regular display for any other optical element, except opd_vmax default changed"
        poppy.AnalyticOpticalElement.display(self,*args, opd_vmax=opd_vmax, **kwargs)


class NIRISS_CLEARP(poppy.CompoundAnalyticOptic):
    """NIRISS 'CLEARP' pupil, including PAR obscuration

        CLEARP pupil info from:
           MODIFIED CALIBRATION OPTIC HOLDER - NIRISS
           DRAWING NO 196847  REV 0  COMDEV
           Design file name 196847Rev0.pdf sent by Loic Albert

        Properties:
          39 mm outer diam, corresponds to the circumscribing pupil of JWST
          2.0 mm vane width
          6.0 mm radius for central obstruction

        Note the circumscribing pupil of JWST is 6603.464 mm in diameter
        (Ball SER on geometric optics model: BALL-JWST-SYST-05-003)
        and therefore the NIRISS pupil magnification is 6.603464/39.0 
        = 0.1693 meters (JWST primary) per mm (NIRISS internal pupil)

        Pupil distortions are not included in this model.

    """
    def __init__(self):
        # CLEARP pupil info from:
        #   MODIFIED CALIBRATION OPTIC HOLDER - NIRISS
        #   DRAWING NO 196847  REV 0  COMDEV
        #   Design file name 196847Rev0.pdf sent by Loic Albert
        # Properties:
        #  39 mm outer diam, corresponds to the circumscribing pupil of JWST
        #  2.0 mm vane width
        #  6.0 mm radius for central obstruction
        # Note the circumscribing pupil of JWST is 6603.464 mm in diameter
        #  (Ball SER on geometric optics model: BALL-JWST-SYST-05-003)

        pupil_mag = 6.603464/39.0
        poppy.CompoundAnalyticOptic.__init__( self, (
                poppy.SecondaryObscuration( secondary_radius = 6.0*pupil_mag,
                                                      support_width = 2.0*pupil_mag,
                                                      n_supports = 3,
                                                      support_angle_offset=90), # align first support with +V2 axis
                poppy.CircularAperture( radius = 39 * pupil_mag /2) ), name = 'CLEARP')






from webbpsf import wfirst

def test_wfirstimager_psf():
    """
    Just test that instantiating WFIRSTImager works and can compute a PSF without raising
    any exceptions
    """
    wi = wfirst.WFIRSTImager()
    wi.calcPSF()



# #!/usr/bin/env python
# import os, sys
# sys.path.append('../lib')
# import numpy as N
# import matplotlib.pyplot as P
# #import matplotlib
# #import aplpy, atpy
# #import RO.DS9
# import logging
# _log = logging.getLogger('t')
# import time
# import itertools
# 
# import webbpsf
# 
# import pysynphot
# from pysynpol import synplot
# 
# 
# def barOutline(xdata, ydata, endstozero=True, *args, **kwargs):
#     """ Draw an outlined-box type plot (like IDL's linestyle=10)
#     Based on http://www.scipy.org/Cookbook/Matplotlib/UnfilledHistograms
# 
#     Parameters
#     ----------
#     xdata : array_like
#         left edge of bin
#     ydata : array_like
#         value for bin
#     endstozero : bool
#         Connect the ends of the histogram to the X axis? default True
# 
# 
#     You can also supply any number of additional matplotlib keywords that
#     will be passed along to plot()
#     """
# 
#     #(histIn, binsIn) = np.histogram(dataIn, *args, **kwargs)
# 
#     stepSize = xdata[1] - xdata[0]
# 
#     bins = N.zeros(len(xdata)*2 + 2, dtype=N.float)
#     data = N.zeros(len(xdata)*2 + 2, dtype=N.float)
#     for bb in range(len(xdata)):
#         bins[2*bb + 1] = xdata[bb]
#         bins[2*bb + 2] = xdata[min(bb+1, len(xdata)-1)]
#         if bb < len(ydata):
#             data[2*bb + 1] = ydata[bb]
#             data[2*bb + 2] = ydata[bb]
# 
#     bins[0] = bins[1]
#     bins[-1] = bins[-2]
#     data[0] = data[1]
#     data[-1] = data[-2]
# 
#     if endstozero:
#         data = N.concatenate(([0], data,[0]))
#         bins = N.concatenate(([bins[0]], bins, [bins[-1]]))
# 
#     P.plot(bins,data, *args, **kwargs)
# 
# 
# 
# def get_rel_fluxes(obsname='miri,im,f560w', nlambda=10, plot=True, verbose=False):
#     raise NotImplemented("Obsolete!")
# 
#     star = pysynphot.Icat('ck04models',7700,0.0,2.0)
#     band0 = pysynphot.ObsBandpass(obsname)
#     #band = band0 *0.01  # convert from percent to real throughput???
#     band=band0
#        
#     # choose reasonable min and max wavelengths
#     w_above10 = N.where(band.throughput > 0.25*band.throughput.max())
#     minwave = band.wave[w_above10].min()  
#     maxwave = band.wave[w_above10].max()
# 
#     wavesteps =  N.linspace(minwave,maxwave,nlambda)
#     deltawave = wavesteps[1]-wavesteps[0]
#     effstims = []
#     t0= time.time()
#     for wave in wavesteps:
#         box = pysynphot.Box(wave, deltawave)
#         obs = pysynphot.Observation(star, band*box)
#         effstims.append(obs.effstim('counts'))
# 
# 
#         if verbose: 
#             _log.info("Box [%.3f-%.3f]: %f cts" % (wave-deltawave/2, wave+deltawave/2, effstims[-1]))
#         #print band.waveunits.Convert(wave,'micron'), effstims[-1]
#     t1 = time.time()
#     print "  that took %f seconds for %d wavelengths" % (t1-t0, nlambda)
# 
# 
#     effstims = N.array(effstims)
#     effstims /= effstims.sum()
# 
#     wave_um =  band.waveunits.Convert(wavesteps,'micron')
#     if plot:
#         P.clf()
#         ax=P.subplot(311)
#         synplot(star)
#         ax.text(0.5, 0.8, "Some random star", horizontalalignment='center', transform = ax.transAxes)
# 
# 
#         ax=P.subplot(312,sharex=ax)
#         synplot(band)
#         ax.text(0.5, 0.8, obsname, horizontalalignment='center', transform = ax.transAxes)
#         if 'fnd' in obsname: ax.set_ybound((0,0.002))
#      
#         ax.set_xbound(1,30)
#         P.draw()
# 
# 
#         P.subplot(313,sharex=ax)
#         #P.plot(wave_um, effstims)
#         barOutline(wave_um, effstims)
#         ax.set_xbound(1,30)
#         
#     #print band.waveunits.Convert(minwave,'micron')
#     #print band.waveunits.Convert(maxwave,'micron')
#     return wave_um, effstims
# 
# 
# def plot_miri_comparison():
# 
#     inst = webbpsf.MIRI()
#     filtlist_W = [f for f in inst.filter_list if f[-1] == 'W']
#     filtlist_C = [f for f in inst.filter_list if f[-1] != 'W']
# 
#     from matplotlib.backends.backend_pdf import PdfPages
#     pdf=PdfPages('weights_miri_comparison.pdf')
# 
# 
#     for filts in [filtlist_W, filtlist_C]:
# 
#         try:
#             os.unlink('/Users/mperrin/software/webbpsf/data/MIRI/filters')
#         except: 
#             pass
#         os.symlink('/Users/mperrin/software/webbpsf/data/MIRI/real_filters', '/Users/mperrin/software/webbpsf/data/MIRI/filters')
#         plotweights('miri', filtlist=filts)
# 
#         os.unlink('/Users/mperrin/software/webbpsf/data/MIRI/filters')
#         os.symlink('/Users/mperrin/software/webbpsf/data/MIRI/fake_filters', '/Users/mperrin/software/webbpsf/data/MIRI/filters')
#         plotweights('miri', filtlist=filts, overplot=True, ls='--')
#         P.draw()
#         pdf.savefig()
# 
#     pdf.close()
# 
# 
# 
# def image_miri_comparison(filter_=None, all=False):
#     inst = webbpsf.MIRI()
# 
#     if all:
#         from matplotlib.backends.backend_pdf import PdfPages
#         pdf=PdfPages('images_miri_comparison.pdf')
#         for f in inst.filter_list:
#             if f[0] != 'F': continue
#             P.clf()
#             image_miri_comparison(f)
#             pdf.savefig()
#         pdf.close()
# 
# 
# 
# 
#     inst.filter = filter_
# 
#     P.subplots_adjust(left=0.02, right=0.98, top=0.99, bottom=0.01)
# 
#     nlambda = inst._filter_nlambda_default[filter_]*2
#     #nlambda=3
#     try:
#         os.unlink('/Users/mperrin/software/webbpsf/data/MIRI/filters')
#     except: 
#         pass
#     os.symlink('/Users/mperrin/software/webbpsf/data/MIRI/real_filters', '/Users/mperrin/software/webbpsf/data/MIRI/filters')
#     im_real = inst.calcPSF(oversample=4, fov_arcsec=15, nlambda=nlambda)
# 
#     os.unlink('/Users/mperrin/software/webbpsf/data/MIRI/filters')
#     os.symlink('/Users/mperrin/software/webbpsf/data/MIRI/fake_filters', '/Users/mperrin/software/webbpsf/data/MIRI/filters')
#     im_fake = inst.calcPSF(oversample=4, fov_arcsec=15, nlambda=nlambda)
# 
#     P.clf()
#     P.subplot(141)
#     webbpsf.display_PSF(im_real,  title='real filter', colorbar_orientation='horizontal')
#  
#     fwhm = webbpsf.measure_fwhm(im_real)
#     ee_real = webbpsf.measure_EE(im_real)
#     P.gca().set_xlabel('FWHM = %.3f", EE(1.0)=%.3f' % (fwhm, ee_real(1.0)))
# 
#     P.subplot(142)
#     webbpsf.display_PSF(im_real, title='box filter from specs', colorbar_orientation='horizontal')
# 
#     fwhm = webbpsf.measure_fwhm(im_fake)
#     ee_fake = webbpsf.measure_EE(im_fake)
#     P.gca().set_xlabel('FWHM = %.3f", EE(1.0)=%.3f' % (fwhm, ee_fake(1.0)))
# 
#  
#     P.subplot(143)
#     webbpsf.display_PSF_difference(im_real, im_fake, title='real filters-box filters', colorbar_orientation='horizontal', print_=True, vmax=1e-5)
# 
#     P.subplot(144)
#     webbpsf.display_PSF_difference(im_real, im_fake, normalize=True, title='real filters-box filters (normalized)', colorbar_orientation='horizontal', vmax=1.0, print_=True)
#    
#     P.gcf().suptitle('MIRI '+filter_, fontsize=20)
# 
# 
# 
# def plot_weights_miri():
#     inst = webbpsf.MIRI()
#     filtlist_W = [f for f in inst.filter_list if f[-1] == 'W']
#     filtlist_C = [f for f in inst.filter_list if f[-1] != 'W']
# 
#     #filtlist_C = [f for f in inst.filter_list if f[-1] == 'C']
# 
# 
#     from matplotlib.backends.backend_pdf import PdfPages
#     pdf=PdfPages('weights_miri.pdf')
# 
#     plotweights('miri', filtlist=filtlist_W)
#     Mstar = pysynphot.Icat('ck04models',3500,0.0,2.0)
#     plotweights('miri', filtlist=filtlist_W, overplot=True, ls='--', source=Mstar)
#     pdf.savefig()
# 
# 
#     plotweights('miri', filtlist=filtlist_C)
#     Mstar = pysynphot.Icat('ck04models',3500,0.0,2.0)
#     plotweights('miri', filtlist=filtlist_C, overplot=True, ls='--', source=Mstar)
#     pdf.savefig()
# 
#     pdf.close()
# 
# 
# def plot_weights_nircam():
#     inst = webbpsf.NIRCam()
#     filtlist_W = [f for f in inst.filter_list if f[-1] == 'W']
#     filtlist_M = [f for f in inst.filter_list if f[-1] == 'M']
#     filtlist_N = [f for f in inst.filter_list if f[-1] == 'N' or f[-1] =='2']
# 
#     #filtlist_C = [f for f in inst.filter_list if f[-1] == 'C']
# 
# 
#     from matplotlib.backends.backend_pdf import PdfPages
#     pdf=PdfPages('weights_nircam.pdf')
# 
#     plotweights('nircam', filtlist=filtlist_W)
#     Mstar = pysynphot.Icat('ck04models',3500,0.0,2.0)
#     plotweights('nircam', filtlist=filtlist_W, overplot=True, ls='--', source=Mstar)
#     pdf.savefig()
# 
# 
#     plotweights('nircam', filtlist=filtlist_M)
#     Mstar = pysynphot.Icat('ck04models',3500,0.0,2.0)
#     plotweights('nircam', filtlist=filtlist_M, overplot=True, ls='--', source=Mstar)
#     pdf.savefig()
# 
#     plotweights('nircam', filtlist=filtlist_N)
#     Mstar = pysynphot.Icat('ck04models',3500,0.0,2.0)
#     plotweights('nircam', filtlist=filtlist_N, overplot=True, ls='--', source=Mstar)
#     pdf.savefig()
# 
# 
#     pdf.close()
# 
# 
# 
# 
# 
# def plotweights(instrument='miri', nlambda =None, filtlist = None, source=None, overplot=False, **kwargs):
# 
#     colorcycle = itertools.cycle(['magenta', 'purple', 'blue', 'skyblue', 'green', 'gold', 'orange', 'red','black'])
# 
#     if source is None:
#         source= pysynphot.Icat('ck04models',5700,0.0,2.0)
# 
#     inst = webbpsf.Instrument(instrument)
#     if filtlist is None:
#         filters = inst.filter_list
#     else:
#         filters = filtlist
# 
#     if not overplot:
#         P.clf()
#         P.subplots_adjust(hspace=0.02)
# 
#     ax=P.subplot(311)
#     #ax.set_color_cycle(colorcycle)
#     synplot(source, **kwargs)
#     ax.set_ylim(1e-0, 1e8)
#     legend_font = matplotlib.font_manager.FontProperties(size=10)
#     P.legend(loc='lower left', prop=legend_font)
# 
#     for filt in filters:
#         color = colorcycle.next()
#         if filt[0].upper() != 'F': continue # skip the IFUs...
#         nlambda = inst._filter_nlambda_default[filt]
#         print "----------- %s: %d -----------" % (filt, nlambda)
#         obsname = ('%s,im,%s'%(instrument, filt)).lower()
#         ax2 = P.subplot(312, sharex=ax)
#         ax2.set_ybound(0,1.1)
#         inst.filter=filt
#         #try:
#         band = inst._getSynphotBandpass() #pysynphot.ObsBandpass(obsname)
#         synplot(band, color=color, **kwargs)
#         ax2.set_ybound(0,1.1)
#         #P.draw()
#             #ls ="-"
#         #except:
#             #P.plot([0,30],[0,0]) # plot something to keep color progression in sync...
#             #ls ="--"
#             #continue
# 
#         #waves, weights = get_rel_fluxes(obsname, nlambda, plot=False)
#         waves, weights = inst._getWeights(source, nlambda=nlambda)
#         #print waves
#         #print weights
# 
#         ax3=P.subplot(313, sharex=ax)
#         P.ylabel("Weight")
#         P.xlabel("Wavelength [$\mu$m]")
# 
#         # barOutline wants to be called with the left edge of each bin, not the center.
#         wave_step = waves[1]-waves[0]
#         plot_waves = N.concatenate( (waves,[waves[-1]+wave_step]))-wave_step/2
#         plot_weights = N.concatenate((weights, [0]))
#         barOutline(plot_waves*1e6, plot_weights, label=filt if not overplot else None, color=color, **kwargs)
# 
#         P.draw()
#         #stop()
#  
#     locs = {'miri': 'upper left', 'nircam': 'upper right', 'tfi': 'upper right'}
#     bounds = {'miri': (4, 30), 'nircam': (0.5, 6), 'tfi': (0.5, 6), 'nirspec': (0.5, 6), 'fgs': (0.5, 6)}
# 
# 
#     P.legend(loc=locs[instrument], prop=legend_font)
#     if instrument =='nircam':
#         ax.set_xscale('linear')
#     ax.set_xbound(*bounds[instrument])
# 
# 
# if __name__ == "__main__":
#     logging.basicConfig(level=logging.INFO, format='%(name)-12s: %(levelname)-8s %(message)s',)
# 
#     #get_rel_fluxes()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This sphinx extension builds off of `sphinx.ext.autosummary` to
clean up some issues it presents in the Astropy docs.

The main issue this fixes is the summary tables getting cut off before the
end of the sentence in some cases.

"""
import re

from sphinx.ext.autosummary import Autosummary

# used in AstropyAutosummary.get_items
_itemsummrex = re.compile(r'^([A-Z].*?\.(?:\s|$))')


class AstropyAutosummary(Autosummary):
    def get_items(self, names):
        """Try to import the given names, and return a list of
        ``[(name, signature, summary_string, real_name), ...]``.
        """
        from sphinx.ext.autosummary import (get_import_prefixes_from_env,
            import_by_name, get_documenter, mangle_signature)

        env = self.state.document.settings.env

        prefixes = get_import_prefixes_from_env(env)

        items = []

        max_item_chars = 50

        for name in names:
            display_name = name
            if name.startswith('~'):
                name = name[1:]
                display_name = name.split('.')[-1]

            try:
                import_by_name_values = import_by_name(name, prefixes=prefixes)
            except ImportError:
                self.warn('[astropyautosummary] failed to import %s' % name)
                items.append((name, '', '', name))
                continue

            # to accommodate Sphinx v1.2.2 and v1.2.3
            if len(import_by_name_values) == 3:
                real_name, obj, parent = import_by_name_values
            elif len(import_by_name_values) == 4:
                real_name, obj, parent, module_name = import_by_name_values

            # NB. using real_name here is important, since Documenters
            #     handle module prefixes slightly differently
            documenter = get_documenter(obj, parent)(self, real_name)
            if not documenter.parse_name():
                self.warn('[astropyautosummary] failed to parse name %s' % real_name)
                items.append((display_name, '', '', real_name))
                continue
            if not documenter.import_object():
                self.warn('[astropyautosummary] failed to import object %s' % real_name)
                items.append((display_name, '', '', real_name))
                continue

            # -- Grab the signature

            sig = documenter.format_signature()
            if not sig:
                sig = ''
            else:
                max_chars = max(10, max_item_chars - len(display_name))
                sig = mangle_signature(sig, max_chars=max_chars)
                sig = sig.replace('*', r'\*')

            # -- Grab the summary

            doc = list(documenter.process_doc(documenter.get_doc()))

            while doc and not doc[0].strip():
                doc.pop(0)
            m = _itemsummrex.search(" ".join(doc).strip())
            if m:
                summary = m.group(1).strip()
            elif doc:
                summary = doc[0].strip()
            else:
                summary = ''

            items.append((display_name, sig, summary, real_name))

        return items


def setup(app):
    # need autosummary, of course
    app.setup_extension('sphinx.ext.autosummary')
    # this replaces the default autosummary with the astropy one
    app.add_directive('autosummary', AstropyAutosummary)



#!/usr/bin/env python
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import astropy.io.fits as fits

from threading import Thread

from . import config
from . import conf

__doc__ = """
Graphical Interface for WebbPSF 

Developed in wxpython by Marshall Perrin

Some code based on/borrowed from Scamp by Klaus Pontoppidan

"""

try:
    import wx, wx.html
except ImportError:
    raise ImportError,"The wxPython module is required to run this program."


import logging
import logging
_log = logging.getLogger('webbpsf')



def _default_options():
    import poppy
    return {'force_coron': False, 'no_sam': False, 'parity':'Either',
                'psf_scale':'log', 'psf_normalize':'Peak', 
                'psf_cmap_str': 'Jet (blue to red)', 'psf_cmap': matplotlib.cm.jet,
                'psf_vmin':1e-4, 'psf_vmax':1.0, 'monochromatic': False, 'fov_in_arcsec': True,
                'parallelization': poppy.conf.use_multiprocessing }



try:
    import pysynphot
    _HAS_PYSYNPHOT=True
except:
    _HAS_PYSYNPHOT=False


import poppy
import webbpsf_core

class WebbPSF_GUI(wx.Frame):
    """ A GUI for the PSF Simulator 

    Documentation TBD!

    """
    def __init__(self, parent=None, id=-1, title="WebbPSF: JWST PSF Calculator"): 
        wx.Frame.__init__(self,parent,id=id,title=title)
        self.parent=parent
        opdserver=None
        self.log_window=None
        # init the object and subobjects
        self.instrument = {}
        self.widgets = {}
        self.vars = {}
        self.advanced_options = _default_options()
        insts = ['NIRCam', 'NIRSpec','NIRISS', 'MIRI', 'FGS']
        for i in insts:
            self.instrument[i] = webbpsf_core.Instrument(i)

        self.inst = self.instrument['NIRCam'] # default

        #invoke link to ITM server if provided?
        if opdserver is not None:
            self._enable_opdserver = True
            self._opdserver = opdserver
        else:
            self._enable_opdserver = False

        
        # create widgets & run
        self._create_widgets_wx()

        # Set up event handler for any worker thread results
        EVT_RESULT(self,self.OnCalcDone)

        self.Show(True)

    def _add_labeled_dropdown(self, name, parent, parentsizer, label="Entry:", choices=None, default=0, width=None, position=(0,0), columnspan=1, expand=True, **kwargs):
        "convenient wrapper for adding a Combobox"

        mylabel = wx.StaticText(parent, -1,label=label)
        parentsizer.Add( mylabel, position,  (1,1), wx.EXPAND|wx.ALIGN_CENTER_VERTICAL)

        if choices is None:
            try:
                choices = self.values[name]
            except:
                choices = ['Option A','Option B']

        if isinstance(default,int):
            value = choices[default]
        else:
            value=default

        #size = (width,25) if width is not None else None
        size=None

        mycombo = wx.ComboBox(parent, -1,value=value, choices=choices, style=wx.CB_DROPDOWN|wx.CB_READONLY, size=size)
        style = wx.EXPAND if expand else 0
        parentsizer.Add( mycombo, (position[0],position[1]+1),  (1,columnspan), style)

        self.widgets[name] = mycombo
 

    def _add_labeled_entry(self, name, parent, parentsizer, label="Entry:", value=None, format="%.2g", 
            width=None, position=(0,0), postlabel=None, **kwargs):
        "convenient wrapper for adding an Entry"
        mylabel = wx.StaticText(parent, -1,label=label)
        parentsizer.Add( mylabel, position,  (1,1), wx.EXPAND)
        self.widgets[name+"_pre_label"] = mylabel



        if value is None:
            try:
                value=format % self.input_options[name]
            except:
                value=""

        size = (width,23) if width is not None else None

        mytext = wx.TextCtrl(parent, -1,value=value, size=size)
        parentsizer.Add( mytext, (position[0],position[1]+1),  (1,1), wx.EXPAND)

        self.widgets[name] = mytext

        if postlabel is not None:
            mylabel2 = wx.StaticText(parent, -1,label=postlabel)
            parentsizer.Add( mylabel2, (position[0],position[1]+2),  (1,1), wx.EXPAND)
            self.widgets[name+"_post_label"] = mylabel2


    def _create_widgets_wx(self):
        """Create a nice GUI using the enhanced widget set provided by 
        the ttk extension to Tkinter, available in Python 2.7 or newer
        """
        #---- create the GUIs
        insts = ['NIRCam', 'NIRSpec','NIRISS', 'MIRI',  'FGS']
        self.sb =self.CreateStatusBar()

        menuBar = WebbPSFMenuBar(self)
        self.SetMenuBar(menuBar)

        self.Bind(wx.EVT_MENU, self.ev_AboutBox, id=wx.ID_ABOUT )
        self.Bind(wx.EVT_MENU, self.ev_Preferences, id=menuBar.ids['preferences'] )
        self.Bind(wx.EVT_MENU, self.ev_SaveAs, id=menuBar.ids['save_psf'] )
        self.Bind(wx.EVT_MENU, self.ev_SaveProfiles, id=menuBar.ids['save_profile'] )
        self.Bind(wx.EVT_MENU, self.ev_calcPSF, id=menuBar.ids['calcPSF'] )
        self.Bind(wx.EVT_MENU, self.ev_options, id=menuBar.ids['calc_options'] )
        self.Bind(wx.EVT_MENU, self.ev_showDocs, id=menuBar.ids['documentation'] )
        self.Bind(wx.EVT_MENU, self.ev_plotspectrum, id=menuBar.ids['display_spectrum'] )
        self.Bind(wx.EVT_MENU, self.ev_displayOptics, id=menuBar.ids['display_optics'] )
        self.Bind(wx.EVT_MENU, self.ev_displayOPD, id=menuBar.ids['display_opd'] )
        self.Bind(wx.EVT_MENU, self.ev_displayPSF, id=menuBar.ids['display_psf'] )
        self.Bind(wx.EVT_MENU, self.ev_displayProfiles, id=menuBar.ids['display_profiles'] )


        top_panel = wx.Panel(self)


        topSizer = wx.BoxSizer(wx.VERTICAL)


        #frame = ttk.Frame(self.root)
        #frame = ttk.Frame(self.root, padx=10,pady=10)

        #ttk.Label(frame, text='James Webb PSF Calculator' ).grid(row=0)

        #===== source properties panel ======
        sb1 = wx.StaticBox(top_panel, label='Source Properties')
        sb1Sizer = wx.StaticBoxSizer(sb1,wx.VERTICAL)


        if _HAS_PYSYNPHOT:
            spectrumPanel = wx.Panel(top_panel)
            spectrumSizer = wx.GridBagSizer()
            
            try:
                choices = poppy.specFromSpectralType("",return_list=True)
                default='G0V'
            except:
                choices = ['Error: $PYSYN_CDBS does not have any spectral models']
                default = choices[0]

            self._add_labeled_dropdown("SpType", spectrumPanel,spectrumSizer, label='    Spectral Type:     ', 
                    choices=choices, default=default, 
                    position=(0,0))
            self.ButtonPlotSpec = wx.Button(spectrumPanel, label='Plot Spectrum')
            self.Bind(wx.EVT_BUTTON, self.ev_plotspectrum, self.ButtonPlotSpec)
            spectrumSizer.Add(self.ButtonPlotSpec, (0,3),(1,1))
            spectrumSizer.AddGrowableCol(2)

            spectrumPanel.SetSizerAndFit(spectrumSizer)
            sb1Sizer.Add(spectrumPanel, 0, wx.ALL|wx.EXPAND, 2)

        posPanel = wx.Panel(top_panel)
        posSizer = wx.GridBagSizer()
        r=0
        self._add_labeled_entry("source_off_r", posPanel,posSizer, label='    Source Position: r=', value='0.0', width=60,  position=(r,0))
        self._add_labeled_entry("source_off_theta", posPanel,posSizer, label='arcsec,  PA=', value='0', width=60, position=(r,2))
        posPanel.SetSizerAndFit(posSizer)
        sb1Sizer.Add(posPanel,0, wx.ALL|wx.EXPAND, 2)


        #===== instruments panels ========
        nb = wx.Notebook(top_panel)
        for iname in insts:
            inst_panel = wx.Panel(nb)
            inst_panel.WebbPSFInstrumentName = iname
            panelSizer = wx.GridBagSizer()


            panelSizer.Add(wx.StaticText(inst_panel, label='Configuration Options for '+iname+"     "), (0,0),(1,3), flag=wx.ALIGN_LEFT)

            instButton = wx.Button(inst_panel, label='Display Optics')
            panelSizer.Add(instButton, (0,4),(1,2), flag=wx.ALIGN_RIGHT)
            self.Bind(wx.EVT_BUTTON, self.ev_displayOptics, instButton)


            self._add_labeled_dropdown(iname+"_filter", inst_panel,panelSizer, label='    Filter:', choices=self.instrument[iname].filter_list, 
                default=self.instrument[iname].filter,  position=(1,0))

            if len(self.instrument[iname].image_mask_list) >0 :
                masks = self.instrument[iname].image_mask_list
                masks.insert(0, "")

                label = '    Slit:    ' if iname=='NIRSpec' else '    Image Stop:'
                self._add_labeled_dropdown(iname+"_coron", inst_panel,panelSizer, label=label, choices=masks,  position=(2,0))


            if len(self.instrument[iname].image_mask_list) >0 :
                masks = self.instrument[iname].pupil_mask_list
                masks.insert(0, "")
                self._add_labeled_dropdown(iname+"_pupil", inst_panel,panelSizer, label='    Pupil Stop:', choices=masks,  position=(3,0))

                fr2 = wx.Panel(inst_panel) 
                fr2Sizer = wx.GridBagSizer()
                self._add_labeled_entry(iname+"_pupilshift_x", fr2,fr2Sizer, label='  pupil shift in X:', value='0', width=40, position=(0,4))
                self._add_labeled_entry(iname+"_pupilshift_y", fr2,fr2Sizer, label=' Y:', value='0', width=40, position=(0,6))

                fr2Sizer.Add(wx.StaticText(fr2, label='% of pupil,' ), (0,8), (1,1))
                self._add_labeled_entry(iname+"_pupil_rot", fr2,fr2Sizer, label=' rotated', value='0', width=40, position=(0,9))
                fr2Sizer.Add(wx.StaticText(fr2, label='deg' ), (0,11), (1,1))
                fr2.SetSizerAndFit(fr2Sizer)

                panelSizer.Add(fr2, (3,3),(1,3))

            

            panelSizer.Add(wx.StaticText(inst_panel, label='Configuration Options for the Telescope (OTE) '), (5,0),(1,5), flag=wx.ALIGN_LEFT)
            opdPanel = wx.Panel(inst_panel)
            opdSizer = wx.GridBagSizer()
            opd_list = self.instrument[iname].opd_list
            opd_list.insert(0,"Zero OPD (perfect)")
            if self._enable_opdserver:
                opd_list.append("OPD from ITM Server")
            default_opd = self.instrument[iname].pupilopd if self.instrument[iname].pupilopd is not None else "Zero OPD (perfect)"

            self._add_labeled_dropdown(iname+"_opd", opdPanel,opdSizer, label='    OPD File:', choices=opd_list, default=default_opd, width=220, position=(0,0), expand=False)
            self.Bind(wx.EVT_COMBOBOX, self.ev_update_OPD_labels, self.widgets[iname+"_opd"])

            self._add_labeled_dropdown(iname+"_opd_i", opdPanel,opdSizer, label=' # ', choices= [str(i) for i in range(10)], width=3, position=(0,2), expand=False)

            self.widgets[iname+"_opd_label"] = wx.StaticText(opdPanel, label=' 0 nm RMS                                            ', style=wx.ALIGN_LEFT|wx.ST_NO_AUTORESIZE)
            opdSizer.Add(self.widgets[iname+"_opd_label"], (0,5),(1,1))
            opdSizer.AddGrowableCol(5)

            instDispButton = wx.Button(opdPanel,label='Display OPD')
            opdSizer.Add(instDispButton, (0,6),(1,1), flag=wx.ALIGN_RIGHT)
            self.Bind(wx.EVT_BUTTON, self.ev_displayOPD, instDispButton)

            opdPanel.SetSizerAndFit(opdSizer)

            panelSizer.Add(opdPanel, (6,0), (1,6), flag=wx.EXPAND|wx.ALL)


            #spacerPanel = wx.Panel(inst_panel)
            #panelSizer.Add(spacerPanel, (7,0), (1,6), flag=wx.EXPAND|wx.ALL)

            panelSizer.AddGrowableCol(2)
            panelSizer.AddGrowableRow(4)
            panelSizer.AddGrowableRow(6)

            inst_panel.SetSizerAndFit(panelSizer)
        
            nb.AddPage(inst_panel, iname)

        self.widgets['tabset'] = nb
        self.ev_update_OPD_labels(None)

        #===== Calculation Options ======

        sb3 = wx.StaticBox(top_panel, label='Calculation Options')
        sb3Sizer = wx.StaticBoxSizer(sb3,wx.VERTICAL)

        calcPanel = wx.Panel(top_panel)
        calcSizer = wx.GridBagSizer()


        r=0 
        self._add_labeled_entry('FOV', calcPanel,calcSizer, label='Field of View:',  value=str(conf.default_fov_arcsec), postlabel='arcsec/side', position=(0,0))
        r+=1
        self._add_labeled_entry('detector_oversampling', calcPanel,calcSizer, label='Output Oversampling:',  width=3, value=str(conf.default_oversampling), postlabel='x finer than instrument pixels       ', position=(r,0))




        r+=1
        self._add_labeled_entry('fft_oversampling', calcPanel,calcSizer, label='Coronagraph FFT Oversampling:',  width=3, value=str(conf.default_oversampling), postlabel='x finer than Nyquist', position=(r,0))
        r+=1
        self._add_labeled_entry('nlambda', calcPanel,calcSizer, label='# of wavelengths:',  width=3, value='', position=(r,0), postlabel='Leave blank for autoselect')
        r+=1

        self._add_labeled_dropdown("jitter", calcPanel,calcSizer, label='Jitter model:', choices=  ['Just use OPDs', 'Gaussian jitter with 7 mas rms', 'Gaussian jitter with 30 mas rms' ], width=20, position=(r,0), columnspan=2)
        r+=1
        output_options=['Oversampled PSF only', 'Oversampled + Detector Res. PSFs', 'Mock full image from JWST DMS']
        self._add_labeled_dropdown("output_format", calcPanel,calcSizer, label='Output Format:', choices=  ['Oversampled image','Detector sampled image','Both as FITS extensions', 'Mock JWST DMS Output' ], width=30, position=(r,0), columnspan=2)

        calcPanel.SetSizerAndFit(calcSizer)
        sb3Sizer.Add(calcPanel,0, wx.ALL|wx.EXPAND, 2)


        #====== button bar ===========

        bbar = wx.Panel(top_panel)
        bbarSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.ButtonCompute = wx.Button(bbar, label='Compute PSF')
        bbarSizer.Add(self.ButtonCompute, 1, flag=wx.ALL|wx.EXPAND, border=3)
        self.ButtonSavePSF = wx.Button(bbar, label='Save PSF As...')
        self.ButtonDisplayPSF = wx.Button(bbar, label='Display PSF')
        self.ButtonDisplayProf =wx.Button( bbar, label='Display Profiles')
        self.ButtonSavePSF.Enable(False)
        self.ButtonDisplayPSF.Enable(False)
        self.ButtonDisplayProf.Enable(False)
        self.ButtonOptions = wx.Button(bbar, label='More Options...')
        self.ButtonQuit = wx.Button(bbar, label='Quit')
        self.widgets['Display PSF'] = self.ButtonDisplayPSF
        self.widgets['Display profiles'] = self.ButtonDisplayProf
        self.widgets['Save PSF As...'] = self.ButtonSavePSF
        bbarSizer.Add(self.ButtonSavePSF , 1, flag=wx.ALL|wx.EXPAND, border=3)
        bbarSizer.Add(self.ButtonDisplayPSF , 1, flag=wx.ALL|wx.EXPAND, border=3)
        bbarSizer.Add(self.ButtonDisplayProf , 1, flag=wx.ALL|wx.EXPAND, border=3)
        bbarSizer.Add(self.ButtonOptions , 1, flag=wx.ALL|wx.EXPAND, border=3)
        bbarSizer.Add(self.ButtonQuit , 1, flag=wx.ALL|wx.EXPAND, border=3)
        bbar.SetSizerAndFit(bbarSizer)

        self.Bind(wx.EVT_BUTTON, self.ev_calcPSF, self.ButtonCompute)
        self.Bind(wx.EVT_BUTTON, self.ev_displayPSF, self.ButtonDisplayPSF)
        self.Bind(wx.EVT_BUTTON, self.ev_displayProfiles, self.ButtonDisplayProf)
        self.Bind(wx.EVT_BUTTON, self.ev_SaveAs, self.ButtonSavePSF)
        self.Bind(wx.EVT_BUTTON, self.ev_options, self.ButtonOptions)
        self.Bind(wx.EVT_BUTTON, self.OnClose, self.ButtonQuit)


        #==== Add items into top sizer of window
        topSizer.Add(sb1Sizer, 5, flag=wx.EXPAND|wx.ALL, border=10)
        topSizer.Add(nb, 10, flag=wx.EXPAND|wx.ALL, border=6)
        topSizer.Add(sb3Sizer, 9, flag=wx.EXPAND|wx.ALL, border=10)
        topSizer.Add(bbar, 0, flag=wx.EXPAND|wx.ALL, border=10)

        top_panel.SetSizerAndFit(topSizer)
        topSizer.Fit(self)
        self.SetSizeHints(self.GetSize().x,self.GetSize().y,-1,-1 ); #can get bigger but not smaller



        self.Bind(wx.EVT_CLOSE, self.OnClose)

        self.Center() 


    def _refresh_window(self):
        """ Force the window to refresh, and optionally to show itself if hidden (for recent matplotlibs)"""
        plt.draw()
        from distutils.version import StrictVersion
        if StrictVersion(matplotlib.__version__) >= StrictVersion('1.1'):
            plt.show(block=False)

    def log(self, messagestring):
        """ Display an informative string in both the log and the window status bar"""
        _log.info(messagestring)
        self.sb.SetStatusText(messagestring)
        self.Refresh()
        self.Update()
        wx.Yield()


    #--- Event Handlers for user-generated events
    def OnClose(self, event):
        dlg = wx.MessageDialog(self,
            "Do you really want to exit WebbPSF?",
            "Confirm Exit", wx.OK|wx.CANCEL|wx.ICON_QUESTION)
        result = dlg.ShowModal()
        dlg.Destroy()
        if result == wx.ID_OK:
            if self.log_window is not None: self.log_window.Destroy()
            self.Destroy()

    def ev_SaveAs(self, event):
        "Event handler for Save As of output PSFs"
        dlg = wx.FileDialog(self, 'Choose Output Filename to Save PSF', defaultDir = os.getcwd(),
                defaultFile ='PSF_%s_%s.fits' %(self.iname, self.filter), 
                wildcard='*.fits',
                style=wx.SAVE)
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
            filename = os.path.abspath(path)
            self.PSF_HDUlist.writeto(filename) 
            self.log("Saved to %s" % filename)
        else:
            self.log("User cancelled save.")

    def ev_SaveProfiles(self,event):
        """ Event handler for Save PSF Profiles"""
        raise NotImplementedError("Not implemented yet")

    def ev_Preferences(self,event):
        """ Event handler for WebbPSF overall preferences"""
        dlg = WebbPSFPreferencesDialog(self)
        results = dlg.ShowModal()
        dlg.Destroy()


    def ev_options(self, event):
        import copy
        import poppy
        oldoptions = copy.copy(self.advanced_options)

        dlg = WebbPSFOptionsDialog(self, input_options = self.advanced_options)
        results = dlg.ShowModal()

        iname = self.widgets['tabset'].GetCurrentPage().WebbPSFInstrumentName

        if dlg.results is not None: # none means the user hit 'cancel'
            self.advanced_options = dlg.results

            if self.advanced_options['monochromatic']:
                self.widgets["nlambda_pre_label"].SetLabel("Wavelength:    ")
                self.widgets["nlambda_post_label"].SetLabel(" [microns]")
            else:
                self.widgets["nlambda_pre_label"].SetLabel("# of wavelengths:")
                self.widgets["nlambda_post_label"].SetLabel("Leave blank for autoselect")

            if self.advanced_options['fov_in_arcsec']:
                self.widgets["FOV_post_label"].SetLabel("arcsec/side")
                if not oldoptions['fov_in_arcsec']: 
                    # we have to convert pixels to arcsec
                    self._setFOV( self._getFOV() * self.instrument[iname].pixelscale) 
            else:
                self.widgets["FOV_post_label"].SetLabel("pixels/side")
                if oldoptions['fov_in_arcsec']: 
                    # we have to convert arcsec to pixels
                    newfov = self._getFOV() / self.instrument[iname].pixelscale
                    if hasattr(newfov, '__len__') and len(newfov) > 1: 
                        newfov = np.asarray(newfov, dtype=int)
                    else:
                        newfov = int(newfov)
                    self._setFOV( newfov)
            poppy.conf.use_multiprocessing = self.advanced_options['parallelization']
 
        dlg.Destroy()


    def ev_showDocs(self,event):
        """ Event handler for show documentation"""
        raise NotImplementedError("Not implemented yet")

    def ev_AboutBox(self, event):
        about = AboutBox()
        about.ShowModal()


    def ev_plotspectrum(self, event):
        "Event handler for Plot Spectrum "
        self._updateFromGUI()
        self.log("Now calculating spectrum model...")

        self.log("Spectral type is "+self.sptype)
        self.log("Selected instrument tab is "+self.iname)
        self.log("Selected instrument filter is "+self.filter)


        plt.clf()

        ax1 = plt.subplot(311)
        spectrum = poppy.specFromSpectralType(self.sptype)
        synplot(spectrum)
        ax1.set_ybound(1e-6, 1e8) # hard coded for now
        ax1.yaxis.set_major_locator(matplotlib.ticker.LogLocator(base=1000))
        legend_font = matplotlib.font_manager.FontProperties(size=10)
        ax1.legend(loc='lower right', prop=legend_font)


        ax2 = plt.subplot(312, sharex=ax1)
        ax2.set_ybound(0,1.1)
        #try:
        band = self.inst._getSynphotBandpass(self.inst.filter) #pysynphot.ObsBandpass(obsname)
        band.name = "%s %s" % (self.iname, self.inst.filter)
        synplot(band) #, **kwargs)
        legend_font = matplotlib.font_manager.FontProperties(size=10)
        plt.legend(loc='lower right', prop=legend_font)

        ax2.set_ybound(0,1.1)


        ax3 = plt.subplot(313, sharex=ax1)
        if self.nlambda is None:
            # Automatically determine number of appropriate wavelengths.
            # Make selection based on filter configuration file
            try:
                #if self.inst.name=='TFI':    # filter names are irrelevant for TFI.
                    #nlambda=5
                #else:
                    #filt_width = self.filter[-1]
                    #lookup_table = {'NIRCam': {'2': 10, 'W':20,'M':3,'N':1}, 
                                    #'NIRSpec':{'W':5,'M':3,'N':1}, 
                                    #'MIRI':{'W':5,'M':3,'N':1}, 
                                    #'FGS':{'W':5,'M':3,'N':1}}

                    #nlambda = lookup_table[self.name][filt_width]
                nlambda = self.inst._filter_nlambda_default[self.filter]
            except:
                nlambda=10
        else:
            nlambda = self.nlambda
        ax1.set_xbound(0.1, 100)
        plt.draw()
        waves, weights = self.inst._getWeights(spectrum, nlambda=nlambda)

        wave_step = waves[1]-waves[0]
        plot_waves = np.concatenate( ([waves[0]-wave_step], waves, [waves[-1]+wave_step])) * 1e6
        plot_weights = np.concatenate(([0], weights,[0]))


        plt.ylabel("Weight")
        plt.xlabel("Wavelength [$\mu$m]")

        ax3.plot(plot_waves, plot_weights,  drawstyle='steps-mid')

        ax1.set_xbound(0.1, 100)

        self._refresh_window()
        self.log("Spectrum displayed")

    def ev_calcPSF(self, event):
        "Event handler for PSF Calculations"
        self._updateFromGUI()
        self.log("Starting PSF calculation...")


        self._refresh_window() # pre-display window for calculation updates as it progresses...
        for w in ['Display PSF', 'Display profiles', 'Save PSF As...']:
           self.widgets[w].Enable(False)

        self.calcthread = PSFCalcThread()
        self.calcthread.runPSFCalc(self.inst, self) 
#        self.PSF_HDUlist = self.inst.calcPSF(source=source, 
#                detector_oversample= self.detector_oversampling,
#                fft_oversample=self.fft_oversampling,
#                fov_arcsec = self.FOV,  nlambda = self.nlambda, display=True)
        #self.PSF_HDUlist.display()


    def OnCalcDone(self, event):
        """ Called when worker thread returns results..."""
        if event.data is None:
            self.log("No result from calculation!")
            self.PSF_HDUlist = None
        else:
            self.PSF_HDUlist = event.data
            for w in ['Display PSF', 'Display profiles', 'Save PSF As...']:
               self.widgets[w].Enable(True)

        self._refresh_window()
        self.log("PSF calculation complete")
        # In either event, the worker is done
        self.calcthread = None


    def ev_displayPSF(self,event):
        "Event handler for Displaying the PSF"
        #self._updateFromGUI()
        #if self.PSF_HDUlist is not None:
        plt.clf()
        poppy.display_PSF(self.PSF_HDUlist, vmin = self.advanced_options['psf_vmin'], vmax = self.advanced_options['psf_vmax'], 
                scale = self.advanced_options['psf_scale'], cmap= self.advanced_options['psf_cmap'], normalize=self.advanced_options['psf_normalize'])
        self._refresh_window()
        self.log("PSF redisplayed")

    def ev_displayProfiles(self,event):
        "Event handler for Displaying the PSF"
        #self._updateFromGUI()
        poppy.display_profiles(self.PSF_HDUlist)        
        self._refresh_window()
        self.log("Radial profiles displayed.")

    def ev_displayOptics(self,event):
        "Event handler for Displaying the optical system"
        self._updateFromGUI()
        self.log("Selected OPD is "+str(self.opd_name))

        plt.clf()
        self.inst.display()
        self._refresh_window()
        self.log("Optical system elements displayed.")

    def ev_displayOPD(self,event):
        import poppy.utils

        self._updateFromGUI()
        if self.inst.pupilopd is None:
            tkMessageBox.showwarning( message="You currently have selected no OPD file (i.e. perfect telescope) so there's nothing to display.", title="Can't Display") 
        else:
            if self._enable_opdserver and 'ITM' in self.opd_name:
                opd = self.inst.pupilopd   # will contain the actual OPD loaded in _updateFromGUI just above
            else:
                opd = fits.getdata(self.inst.pupilopd[0])     # in this case self.inst.pupilopd is a tuple with a string so we have to load it here. 

            if len(opd.shape) >2:
                opd = opd[self.opd_i,:,:] # grab correct slice

            masked_opd = np.ma.masked_equal(opd,  0) # mask out all pixels which are exactly 0, outside the aperture
            cmap = matplotlib.cm.jet
            cmap.set_bad('k', 0.8)

            plt.clf()
            plt.imshow(masked_opd, cmap=cmap, interpolation='nearest', vmin=-0.5, vmax=0.5)
            poppy.utils.imshow_with_mouseover(masked_opd, cmap=cmap, interpolation='nearest', vmin=-0.5, vmax=0.5)
            plt.title("OPD from %s, #%d" %( os.path.basename(self.opd_name), self.opd_i))
            cb = plt.colorbar(orientation='vertical')
            cb.set_label('microns')

            f = plt.gcf()
            plt.text(0.4, 0.02, "OPD WFE = %6.2f nm RMS" % (masked_opd.std()*1000.), transform=f.transFigure)
        self.log("Optical Path Difference (OPD) now displayed.")

        self._refresh_window()

    def ev_launch_ITM_dialog(self, event):
        tkMessageBox.showwarning( message="ITM dialog box not yet implemented", title="Can't Display") 

    def ev_update_OPD_labels(self, event):
        "Update the descriptive text for all OPD files"
        for iname in self.instrument.keys():
            self.ev_update_OPD_label(self.widgets[iname+"_opd"], self.widgets[iname+"_opd_label"], iname)

    def ev_update_OPD_label(self, widget_combobox, widget_label, iname):
        "Update the descriptive text displayed about one OPD file"
        from wx.lib.wordwrap import wordwrap
        showitm=False # default is do not show
        filename = os.path.join( self.instrument[iname]._datapath, 'OPD',  widget_combobox.GetValue() )

        if filename.endswith(".fits"):
            try:
                header_summary = fits.getheader(filename)['SUMMARY']
            except:
                header_summary = 'could not obtain a summary from FITS header'
            #self.widgets[iname+"_opd_i"]['state'] = 'readonly'
        else:  # Special options for non-FITS file inputs
            #self.widgets[iname+"_opd_i"]['state'] = 'disabled'
            val = widget_combobox.GetValue()
            if 'Zero' in val:
                header_summary = " 0 nm RMS"
            elif 'ITM' in val and self._enable_opdserver:
                header_summary= "Get OPD from ITM Server"
                showitm=True
            elif 'ITM' in val and not self._enable_opdserver:
                header_summary = "ITM Server is not running or otherwise unavailable."
            else: # other??
                header_summary = "   "

        widget_label.SetLabel( header_summary  )
        widget_label.Wrap(250)
        #widget_label.SetLabel( wordwrap( header_summary, 40, wx.ClientDC(self))  )


        #if showitm:
        #    self.widgets[iname+"_itm_coords"].grid() # re-show ITM options
        #else:
        #    self.widgets[iname+"_itm_coords"].grid_remove()  # hide ITM options

    def _updateFromGUI(self):
        """ Update the object's state with all the settings from the GUI
        """
        # get GUI values
        if _HAS_PYSYNPHOT:
            self.sptype = self.widgets['SpType'].GetValue()
        self.iname = self.widgets['tabset'].GetCurrentPage().WebbPSFInstrumentName


        if self.advanced_options['monochromatic']:
            # monochromatic mode
            self.nlambda=1
            self.monochromatic_wavelength = float(self.widgets['nlambda'].GetValue()) * 1e-6 # note that the GUI is in microns, so convert to meters here.

            if self.monochromatic_wavelength < 0.6e-6 or self.monochromatic_wavelength > 30e-6:
                _log.error("Invalid wavelength. Please enter a value between 0.6 - 30 microns")
                raise ValueError('Invalid wavelength outside of range 0.6 - 30 microns: {0:f}'.format(self.monochromatic_wavelength))
            try:
                pass
            except:
                _log.error("Could not obtain a wavelength. Please check the value of the wavelength field and try again.")
                raise ValueError('Could not obtain a wavelength from string "{0}"'.format(self.widgets['nlambda'].GetValue()))
        else:
            # normal broadband mode
            self.monochromatic_wavelength = None
            try:
                self.nlambda= int(self.widgets['nlambda'].GetValue())
            except:
                self.nlambda = None # invoke autoselect for nlambda
                



        self.FOV= self._getFOV() #float(self.widgets['FOV'].GetValue())
        self.fft_oversampling= int(self.widgets['fft_oversampling'].GetValue())
        self.detector_oversampling= int(self.widgets['detector_oversampling'].GetValue())

        self.output_type = self.widgets['output_format'].GetValue()

        options = {}
        options['rebin'] = not (self.output_type == 'Oversampled PSF only')  #was downsample, which seems wrong?
        options['mock_dms'] = (self.output_type == 'Mock full image from JWST DMS')
        jitterchoice = self.widgets['jitter'].GetValue()
        if jitterchoice == 'Just use OPDs':
            options['jitter'] = None
        elif jitterchoice == 'Gaussian jitter with 7 mas rms':
            options['jitter'] = 'gauss'
            options['jitter_sigma'] = 0.007
        elif jitterchoice == 'Gaussian jitter with 30 mas rms':
            options['jitter'] = 'gauss'
            options['jitter_sigma'] = 0.030
        else: 
            _log.error("Unknown value for jitter selection: "+jitterchoice)



        # and get the values that may have previously been set by the 'advanced options' dialog
        if self.advanced_options is not None:
            for a in self.advanced_options.keys():
                options[a] = self.advanced_options[a]


        # configure the relevant instrument object
        self.inst = self.instrument[self.iname]
        self.filter = self.widgets[self.iname+"_filter"].GetValue() # save for use in default filenames, etc.
        self.inst.filter = self.filter
        _log.info("Selected filter: "+self.filter)

        self.opd_name = self.widgets[self.iname+"_opd"].GetValue()
        if self._enable_opdserver and 'ITM' in self.opd_name:
            # get from ITM server
            self.opd_i= 0
            self.inst.pupilopd = self._opdserver.get_OPD(return_as="FITS")
            self.opd_name = "OPD from ITM OPD GUI"

        elif self.opd_name == "Zero OPD (perfect)": 
            # perfect OPD
            self.opd_name = "Perfect"
            self.inst.pupilopd = None
        else:
            # Regular FITS file version
            self.opd_name= self.widgets[self.iname+"_opd"].GetValue()
            self.opd_i= int(self.widgets[self.iname+"_opd_i"].GetValue())
            self.inst.pupilopd = (self.inst._datapath+os.sep+"OPD"+os.sep+self.opd_name,self.opd_i)  #filename, slice

        self.log("Selected OPD is "+str(self.opd_name))


        if self.iname+"_coron" in self.widgets:
            self.inst.image_mask = self.widgets[self.iname+"_coron"].GetValue()
            self.inst.pupil_mask = self.widgets[self.iname+"_pupil"].GetValue()
            # TODO read in mis-registration options here.


            options['source_offset_r'] = float(self.widgets["source_off_r"].GetValue())
            options['source_offset_theta'] = float(self.widgets["source_off_theta"].GetValue())
            options['pupil_shift_x'] = float(self.widgets[self.iname+"_pupilshift_x"].GetValue())/100. # convert from percent to fraction
            options['pupil_shift_y'] = float(self.widgets[self.iname+"_pupilshift_y"].GetValue())/100. # convert from percent to fraction

        self.inst.options = options


    def _getFOV(self):
        """ Get field of view, either as a scalar number for square or a 
        2-element ndarray for rectangular 
        
        Note that if it is a 2-element paid, we flip the order of the elements. 
        This is to facilitate a more intuitive "x,y" ordering for the user interface.
        """
        fovstr = self.widgets['FOV'].GetValue()

        try:
            if ',' in fovstr:
                parts = fovstr.split(',')
                return np.asarray( parts[0:2], dtype=float)[::-1]
            else:
                return float(self.widgets['FOV'].GetValue())
        except:
            _log.error("Invalid entry in FOV field. Please check it and try again")
            raise ValueError("Invalid entry in FOV field. Please check it and try again")


    def _setFOV(self, newvalue):
        """ Get field of view, either as a scalar number for square or a 
        2-element ndarray for rectangular 

        Note that if it is a 2-element paid, we flip the order of the elements. 
        This is to facilitate a more intuitive "x,y" ordering for the user interface.
        """
        if hasattr(newvalue, '__iter__'):
            newstring = ",".join( (str(newvalue[1]), str(newvalue[0]) ))
        else:
            newstring = str(newvalue)

        self.widgets['FOV'].SetValue(newstring)

#-------------------------------------------------------------------------
# Class to run the actual PSF calculation in a background thread, to keep the
# GUI still responsive
# This code based on examples at http://wiki.wxpython.org/LongRunningTasks         
# and http://www.blog.pythonlibrary.org/2010/05/22/wxpython-and-threads/

class PSFCalcThread(Thread):
    def __init__(self):
        """Init Worker Thread Class."""
        Thread.__init__(self)
        self.start()    # start the thread

    def runPSFCalc(self,  instrument, masterapp):

        if _HAS_PYSYNPHOT:
            source = poppy.specFromSpectralType(masterapp.sptype)
        else:
            source=None # generic flat spectrum
        print "starting calc in thread"

        if instrument.options['fov_in_arcsec']:
            fov_arcsec = masterapp.FOV
            fov_pixels = None
        else:
            fov_arcsec = None
            fov_pixels = masterapp.FOV
 
        PSF_HDUlist = instrument.calcPSF(source=source, 
                detector_oversample = masterapp.detector_oversampling,
                fft_oversample = masterapp.fft_oversampling,
                fov_arcsec = fov_arcsec, fov_pixels=fov_pixels,  
                nlambda = masterapp.nlambda, 
                monochromatic=masterapp.monochromatic_wavelength, 
                display = True)

        wx.PostEvent(masterapp, ResultEvent(PSF_HDUlist)) # send results back to master thread


# Define notification event for thread completion
EVT_RESULT_ID = wx.NewId()
def EVT_RESULT(win, func):
    """Define Result Event."""
    win.Connect(-1, -1, EVT_RESULT_ID, func)

class ResultEvent(wx.PyEvent):
    """Simple event to carry arbitrary result data."""
    def __init__(self, data):
        """Init Result Event."""
        wx.PyEvent.__init__(self)
        self.SetEventType(EVT_RESULT_ID)
        self.data = data



#-------------------------------------------------------------------------

class WebbPSFMenuBar(wx.MenuBar):
    def __init__(self,parent):
        wx.MenuBar.__init__(self)
        item_keys =['save_psf','save_profile', 'documentation', 'preferences', 'calc_options', 'calcPSF', 
                'display_spectrum', 'display_optics','display_opd', 'display_psf', 'display_profiles']

        self.ids = {}
        for key in item_keys:
            self.ids[key]=wx.NewId()

        # File menu
        filemenu = wx.Menu()
        self.SavePSF = filemenu.Append(self.ids['save_psf'], '&Save PSF as...\tCtrl+Shift+S')
        self.SaveProfile =filemenu.Append(self.ids['save_profile'], 'Save &profile data as...\tCtrl+Shift+P')
        filemenu.AppendSeparator()
        self.Preferences=filemenu.Append(self.ids['preferences'], 'Preferences...\tCtrl+,')
        filemenu.AppendSeparator()
        self.Exit = filemenu.Append(wx.ID_EXIT, 'E&xit', 'Exit this program')

        # these start out disabled since no PSF calculated yet:
        self.SavePSF.Enable(False)
        self.SaveProfile.Enable(False)

        # Edit menu
        editmenu = wx.Menu()
        editmenu.Append(wx.ID_CUT, 'Cut\tCtrl-X')
        editmenu.Append(wx.ID_COPY, 'Copy\tCtrl-C')
        editmenu.Append(wx.ID_PASTE, 'Paste\tCtrl-V')

        #Calculation Menu
        calcmenu = wx.Menu()
        calcmenu.Append(self.ids['calcPSF'], 'Compute PSF')
        self.Preferences=calcmenu.Append(self.ids['calc_options'], 'More Options...')
        calcmenu.AppendSeparator()
        calcmenu.Append(self.ids['display_spectrum'], 'Display Spectrum')
        calcmenu.Append(self.ids['display_optics'], 'Display Optics')
        calcmenu.Append(self.ids['display_opd'], 'Display OPD')
        calcmenu.Append(self.ids['display_psf'], 'Display PSF')
        calcmenu.Append(self.ids['display_profiles'], 'Display PSF Profiles')

        #Help Menu
        helpmenu = wx.Menu()
        self.Docs = helpmenu.Append(self.ids['documentation'], 'WebbPSF Documentation\tCtrl+d')
        self.About = helpmenu.Append(wx.ID_ABOUT, '&About WebbPSF', 'About this program')

        self.Append(filemenu, '&File')
        self.Append(editmenu, '&Edit')
        self.Append(calcmenu, '&Calculation')
        self.Append(helpmenu, '&Help')

#-------------------------------------------------------------------------
class WebbPSFDialog(wx.Dialog):
    """ Generic dialog box for WebbPSF 

    TODO: investigate wx.Validator to validate the text input fields
    """
    def __init__(self, parent=None, id=-1, title="Dialog", **kwargs):
        wx.Dialog.__init__(self,parent,id=id,title=title, **kwargs)

        self.parent=parent

        self.results = None # in case we cancel this gets returned
        self.widgets = {}
        self.values = {}


        #self._createWidgets()
 
    def _add_labeled_dropdown(self, name, parent, parentsizer, label="Entry:", choices=None, default=0, width=5, position=(0,0), columnspan=1, **kwargs):
        """convenient wrapper for adding a Combobox

        columnspan sets the span for the combobox itself
        """

        mylabel = wx.StaticText(parent, -1,label=label)
        parentsizer.Add( mylabel, position,  (1,1), wx.EXPAND)

        if choices is None:
            try:
                choices = self.values[name]
            except:
                choices = ['Option A','Option B']

        if isinstance(default,int):
            value = choices[default]
        else:
            value=default

        mycombo = wx.ComboBox(parent, -1,value=value, choices=choices, style=wx.CB_DROPDOWN|wx.CB_READONLY)
        parentsizer.Add( mycombo, (position[0],position[1]+1),  (1,columnspan), wx.EXPAND)

        self.widgets[name] = mycombo
 

    def _add_labeled_entry(self, name, parent, parentsizer, label="Entry:", value=None, format="%.2g", 
            width=5, position=(0,0), postlabel=None, **kwargs):
        "convenient wrapper for adding an Entry"
        mylabel = wx.StaticText(parent, -1,label=label)
        parentsizer.Add( mylabel, position,  (1,1), wx.EXPAND)

        if value is None:
            try:
                value=format % self.input_options[name]
            except:
                value=""
        else:
            try: 
                value = format % value
            except:
                pass

        mytext = wx.TextCtrl(parent, -1,value=value)
        parentsizer.Add( mytext, (position[0],position[1]+1),  (1,1), wx.EXPAND)

        self.widgets[name] = mytext

        if postlabel is not None:
            mylabel2 = wx.StaticText(parent, -1,label=postlabel)
            parentsizer.Add( mylabel2, (position[0],position[1]+2),  (1,1), wx.EXPAND)


    def _createWidgets(self):
        pass
        # subclass me


    def OnButtonOK(self, event):
        pass


class WebbPSFOptionsDialog(WebbPSFDialog):
    """ Dialog box for WebbPSF options 

    TODO: investigate wx.Validator to validate the text input fields
    """
    def __init__(self, parent=None, id=-1, title="WebbPSF Options", 
            input_options=_default_options()): 
        WebbPSFDialog.__init__(self, parent,id=id,title=title)
        self.input_options = input_options

        colortables = [
         ('Jet (blue to red)',matplotlib.cm.jet),
         ('Gray', matplotlib.cm.gray),
         ('Heat (black-red-yellow)', matplotlib.cm.gist_heat),
         ('Copper (black to tan)',matplotlib.cm.copper),
         ('Stern',matplotlib.cm.gist_stern),
         ('Prism (repeating rainbow)', matplotlib.cm.prism)]

        try:
            import collections
            self.colortables = collections.OrderedDict(colortables)
        except:
            self.colortables = dict(colortables)

        self.values['force_coron'] = ['regular propagation (MFT)', 'full coronagraphic propagation (FFT/SAM)']
        self.values['no_sam'] = ['semi-analytic method if possible', 'basic FFT method always']
        self.values['monochromatic'] = ['Broadband', 'Monochromatic']
        self.values['parallelization'] = ['Sequential', 'Parallelized']
        self.values['fov_in_arcsec'] = ['Arcseconds', 'Pixels']


        self._createWidgets()
 

    def _createWidgets(self):

        topSizer = wx.BoxSizer(wx.VERTICAL)

        topPanel = wx.Panel(self)
        topPanelSizer = wx.FlexGridSizer(rows=3,cols=1, hgap=5, vgap=10)
        #topPanelSizer = wx.BoxSizer(wx.VERTICAL)

        panel1 = wx.Panel(topPanel, style=wx.SIMPLE_BORDER|wx.EXPAND)
        sizer = wx.GridBagSizer()

        txt = wx.StaticText(panel1,-1,label="Propagation Calculation Options")
        sizer.Add(txt,(0,0),(1,3),wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL)


        r=1
        self._add_labeled_dropdown("monochromatic", panel1,sizer, 
                label='    Broadband or monochromatic? ', 
                default = 1 if self.input_options['monochromatic'] else 0, position=(r,0))
        r+=1
        self._add_labeled_dropdown("parallelization", panel1,sizer, 
                label='    Parallelize Wavelengths? ', 
                default = 1 if self.input_options['parallelization'] else 0, position=(r,0))
        r+=1
        self._add_labeled_dropdown("force_coron", panel1,sizer, 
                label='    Direct imaging calculations use: ', 
                default = 1 if self.input_options['force_coron'] else 0, position=(r,0))

        r+=1
        self._add_labeled_dropdown("no_sam", panel1,sizer, 
                label='    Coronagraphic calculations use', 
                default= 1 if self.input_options['no_sam'] else 0, position=(r,0))
        r+=1
        self._add_labeled_dropdown("parity", panel1,sizer, 
                label='    Output pixel grid parity is', 
                choices=['odd', 'even', 'either'], default=self.input_options['parity'], position=(r,0))
        r+=1
        self._add_labeled_dropdown("fov_in_arcsec", panel1,sizer, 
                label='    Specify field of view in: ', 
                default = 0 if self.input_options['fov_in_arcsec'] else 1, position=(r,0))

 
        #sizer.AddGrowableCol(0)
        panel1.SetSizerAndFit(sizer)

        panel2 = wx.Panel(topPanel, style=wx.SIMPLE_BORDER|wx.EXPAND)
        sizer = wx.GridBagSizer()

        r=0
        txt = wx.StaticText(panel2,-1,label="PSF Display Options")
        sizer.Add(txt,(r,0),(1,3),wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL)


        r+=1
        self._add_labeled_dropdown("psf_scale", panel2,sizer, label='    Display scale:', choices=['log','linear'],
                default=self.input_options['psf_scale'], position=(r,0))

        r+=1
        self._add_labeled_entry("psf_vmin", panel2,sizer, label='    Min scale value:', 
            format="%.2g", width=7, 
            position=(r,0))
        r+=1
        self._add_labeled_entry("psf_vmax", panel2,sizer, label='    Max scale value:', 
            format="%.2g", width=7, 
            position=(r,0))
        r+=1
        self._add_labeled_dropdown("psf_normalize", panel2,sizer,
                label='    Normalize PSF to:', choices=['Total', 'Peak'], 
                default=self.input_options['psf_normalize'], position=(r,0))
        r+=1
        self._add_labeled_dropdown("psf_cmap", panel2,sizer, label='    Color table:', 
                choices=[a for a in self.colortables.keys()],  
                default=self.input_options['psf_cmap_str'], 
                position=(r,0))
        panel2.SetSizerAndFit(sizer)


        bbar = self.CreateStdDialogButtonSizer( wx.OK | wx.CANCEL)
        self.Bind(wx.EVT_BUTTON, self.OnButtonOK, id=wx.ID_OK)
        #self.Bind(wx.EVT_BUTTON, self.OnButtonCancel, id=wx.ID_CANCEL)
 

        topPanelSizer.Add(panel1, 1,wx.EXPAND)
        topPanelSizer.Add(panel2, 1, wx.EXPAND)
        #topPanelSizer.Add(panel3,1, wx.EXPAND)
        topPanelSizer.Add(bbar,1, wx.EXPAND)
        #topPanel.AddGrowableCol(0)
        topPanel.SetSizerAndFit(topPanelSizer)
 
        topSizer.Add(topPanel, 1, flag=wx.EXPAND|wx.ALL, border=10)
        self.SetSizerAndFit(topSizer)
        self.Show(True)
    #def OnButtonCancel(self, event):
        #print "User pressed Cancel"
        #self.Close()
        #self.Destroy()

    def OnButtonOK(self, event):
        print "User pressed OK"
        try:
            results = {}
            results['force_coron'] =    self.widgets['force_coron'].GetValue() == 'full coronagraphic propagation (FFT/SAM)'
            results['no_sam'] =         self.widgets['no_sam'].GetValue() == 'basic FFT method always'
            results['monochromatic'] =  self.widgets['monochromatic'].GetValue() == 'Monochromatic'
            results['parallelization'] =  self.widgets['parallelization'].GetValue() == 'Parallelized'
            results['fov_in_arcsec'] =  self.widgets['fov_in_arcsec'].GetValue() == 'Arcseconds'
            results['parity'] =         self.widgets['parity'].GetValue() 
            results['psf_scale'] =      self.widgets['psf_scale'].GetValue() 
            results['psf_vmax'] = float(self.widgets['psf_vmax'].GetValue())
            results['psf_vmin'] = float(self.widgets['psf_vmin'].GetValue())
            results['psf_cmap_str'] =   self.widgets['psf_cmap'].GetValue()
            results['psf_cmap'] =       self.colortables[self.widgets['psf_cmap'].GetValue() ]
            results['psf_normalize'] =  self.widgets['psf_normalize'].GetValue()


            print results
            self.results = results # for access from calling routine

            self.Close()
            #self.Destroy() # return... If called as a modal dialog, should ShowModal and Destroy from calling routine?
        except:
            _log.error("Invalid entries in one or more fields. Please check values and re-enter!")

#-------------------------------------------------------------------------

class WebbPSFPreferencesDialog(WebbPSFDialog):
    """ Dialog box for WebbPSF options 

    TODO: investigate wx.Validator to validate the text input fields
    """
    def __init__(self, parent=None, id=-1, title="WebbPSF Preferences"): 
        WebbPSFDialog.__init__(self, parent,id=id,title=title, size=(800,400))

        self._createWidgets()

    def _createWidgets(self):

        topSizer = wx.BoxSizer(wx.VERTICAL)

        topPanel = wx.Panel(self)
        topPanelSizer = wx.FlexGridSizer(rows=4,cols=1, hgap=5, vgap=10)

        panel1 = wx.Panel(topPanel, style=wx.SIMPLE_BORDER|wx.EXPAND)
        sizer = wx.GridBagSizer()

        txt = wx.StaticText(panel1,-1,label="Directory Paths")
        sizer.Add(txt,(0,0),(1,3),wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL)

        r=1
        self._add_labeled_entry("WEBBPSF_PATH", panel1,sizer, label='    WebbPSF Data Path:', 
            value = str(conf.WEBBPSF_PATH), 
            format="%50s", position=(r,0))

        self.ButtonBrowseDir = wx.Button(panel1, label='Browse...')
        sizer.Add(self.ButtonBrowseDir, (r,4), (1,1), wx.EXPAND)
        self.Bind(wx.EVT_BUTTON, self.ev_browseDataDirectory, self.ButtonBrowseDir)
        panel1.SetSizerAndFit(sizer)



        panel2 = wx.Panel(topPanel, style=wx.SIMPLE_BORDER|wx.EXPAND)
        sizer = wx.GridBagSizer()

        txt = wx.StaticText(panel2,-1,label="Calculation Defaults")
        sizer.Add(txt,(0,0),(1,3),wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL)

        r=1
        self._add_labeled_entry("default_oversampling", panel2,sizer, label='    Default Oversampling:', 
            value = conf.default_oversampling, 
            format="%.2g", width=7, position=(r,0))
        r+=1
        self._add_labeled_entry("default_fov_arcsec", panel2,sizer, label='    Default FOV [arcsec]:', 
            value = conf.default_fov_arcsec, 
            format="%.2g", width=7, position=(r,0))

        # update conf.default_oversampling , default_output_mode, default_fov_arcsec, WEBBPSF_PATH, 
        # use_multiprocessing= True/False 
        # n_processes= #
        # use_fftw = True/False

        panel2.SetSizerAndFit(sizer)

        panel3 = wx.Panel(topPanel, style=wx.SIMPLE_BORDER|wx.EXPAND)
        sizer = wx.GridBagSizer()

        r=0
        txt = wx.StaticText(panel3,-1,label="Fourier Transform Options")
        sizer.Add(txt,(r,0),(1,3),wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL)

        r+=1
        self._add_labeled_dropdown("use_fftw", panel3,sizer, label='    Use FFTW for FFTs:', choices=['True','False'],
                default="True" if poppy.conf.use_fftw else "False", position=(r,0))
        r+=1
        self._add_labeled_dropdown("use_multiprocessing", panel3,sizer, label='    Use Multiprocessing for DFTs:', choices=['True','False'],
                default="True" if poppy.conf.use_multiprocessing else "False", position=(r,0))

        panel3.SetSizerAndFit(sizer)
#
#
#        r+=1
#        self._add_labeled_entry("psf_vmin", panel2,sizer, label='    Min scale value:', 
#            format="%.2g", width=7, position=(r,0))
#        r+=1
#        self._add_labeled_entry("psf_vmax", panel2,sizer, label='    Max scale value:', 
#            format="%.2g", width=7, 
#            position=(r,0))
#        r+=1
#        self._add_labeled_dropdown("psf_normalize", panel2,sizer,
#                label='    Normalize PSF to:', choices=['Total', 'Peak'], 
#                default=self.input_options['psf_normalize'], position=(r,0))
#        r+=1
#        self._add_labeled_dropdown("psf_cmap", panel2,sizer, label='    Color table:', 
#                choices=[a for a in self.colortables.keys()],  
#                default=self.input_options['psf_cmap_str'], 
#                position=(r,0))
#        panel2.SetSizerAndFit(sizer)
#

        bbar = self.CreateStdDialogButtonSizer( wx.OK | wx.CANCEL)
        self.Bind(wx.EVT_BUTTON, self.OnButtonOK, id=wx.ID_OK)

        topPanelSizer.Add(panel1, 1,wx.EXPAND)
        topPanelSizer.Add(panel2, 1, wx.EXPAND)
        topPanelSizer.Add(panel3, 1, wx.EXPAND)
        topPanelSizer.Add(bbar,1, wx.EXPAND)
        topPanel.SetSizerAndFit(topPanelSizer)
 
        topSizer.Add(topPanel, 1, flag=wx.EXPAND|wx.ALL, border=10)
        self.SetSize((800,300))
        self.SetSizerAndFit(topSizer)

        self.Show(True)

  
    def ev_browseDataDirectory(self, event):
        dlg = wx.DirDialog(self, 'Choose WebbPSF Data Directory', defaultPath=os.getcwd())

        if dlg.ShowModal() == wx.ID_OK:
            print "Directory Selected: "+dlg.GetPath()
        else:
            print "Dialog cancelled"

    def OnButtonOK(self, event):
        print "User pressed OK"
        try:
            conf.default_oversampling = int(self.widgets['default_oversampling'].GetValue())
            conf.default_fov_arcsec = float(self.widgets['default_fov_arcsec'].GetValue())

#            results['force_coron'] =    self.widgets['force_coron'].GetValue() == 'full coronagraphic propagation (FFT/SAM)'
#            results['no_sam'] =         self.widgets['no_sam'].GetValue() == 'basic FFT method always'
#            results['monochromatic'] =  self.widgets['monochromatic'].GetValue() == 'Monochromatic'
#            results['fov_in_arcsec'] =  self.widgets['fov_in_arcsec'].GetValue() == 'Arcseconds'
#            results['parity'] =         self.widgets['parity'].GetValue() 
#            results['psf_scale'] =      self.widgets['psf_scale'].GetValue() 
#            results['psf_vmax'] = float(self.widgets['psf_vmax'].GetValue())
#            results['psf_vmin'] = float(self.widgets['psf_vmin'].GetValue())
#            results['psf_cmap_str'] =   self.widgets['psf_cmap'].GetValue()
#            results['psf_cmap'] =       self.colortables[self.widgets['psf_cmap'].GetValue() ]
#            results['psf_normalize'] =  self.widgets['psf_normalize'].GetValue()
#
#
#            print results
#            self.results = results # for access from calling routine

            utils.save_config()
            self.Close()
            #self.Destroy() # return... If called as a modal dialog, should ShowModal and Destroy from calling routine?
        except:
            _log.error("Invalid entries in one or more fields. Please check values and re-enter!")


#-------------------------------------------------------------------------
#
# Classes for displaying log messages in a separate window

class WxLog(logging.Handler):
    def __init__(self, ctrl):
        logging.Handler.__init__(self)
        self.ctrl = ctrl
    def emit(self, record):
        self.ctrl.AppendText(self.format(record)+"\n")
        self.ctrl.Update()

class LogFrame(wx.Frame):
    def __init__(self, parent=None, id=-1, pos=None, size=(600,300)):
        wx.Frame.__init__(self, parent, id=id, title="WebbPSF tasks log", size=size,pos=pos)
        #self.level = 4
        log = wx.TextCtrl(self, style=wx.TE_MULTILINE)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(log, 1, wx.EXPAND)
        self.SetSizer(sizer)


        hdlr = WxLog(log)
        hdlr.setFormatter(logging.Formatter('%(name)-8s %(levelname)-8s: %(message)s'))

        self.hdlr = hdlr # save for later deletion

        logging.getLogger('webbpsf').addHandler(hdlr)
        logging.getLogger('poppy').addHandler(hdlr)
        
    def __del__(self):
        logging.getLogger('webbpsf').removeHandler(self.hdlr)
        logging.getLogger('poppy').removeHandler(self.hdlr)

#-------------------------------------------------------------------------

class HtmlWindow(wx.html.HtmlWindow):
    def __init__(self, parent, id, size=(600,400)):
        wx.html.HtmlWindow.__init__(self,parent, id, size=size)
        if "gtk2" in wx.PlatformInfo:
            self.SetStandardFonts()

    def OnLinkClicked(self, link):
        wx.LaunchDefaultBrowser(link.GetHref())

class AboutBox(wx.Dialog):
    def __init__(self):
        wx.Dialog.__init__(self, None, wx.ID_ANY, "About WebbPSF",
            style=wx.DEFAULT_DIALOG_STYLE|wx.THICK_FRAME|wx.RESIZE_BORDER|
                wx.TAB_TRAVERSAL)


        aboutText = """<p>This is the <b>WebbPSF Point Spread Function Simulator</b> for the James Webb Space Telescope (JWST),
<p>
<center>
Version %(webbpsf)s <P>
See <a href="http://www.stsci.edu/jwst/software/webbpsf">the WebbPSF home page</a>
</center>
<p>
(c) 2010-2013 by Marshall Perrin, <a href="mailto:mperrin@stsci.edu">mperrin@stsci.edu</a>.
<br>
With contributions from: Anand Sivaramakrishnan, Remi Soummer, &amp; Klaus Pontoppidan
<p>
WebbPSF is running with the following software versions:
<ul>
<li><b>Python</b>: %(python)s 
<li><b>numpy</b>: %(numpy)s
<li><b>matplotlib</b>: %(matplotlib)s
<li><b>astropy</b>: %(astropy)s
<li><b>wxPython</b>: %(wxpy)s  
<li><b>pysynphot</b>: %(pysynphot)s  
<li><b>pyFFTW</b>: %(pyfftw)s  
<li><b>PyFFTW3</b>: %(pyfftw3)s  
</ul>
</p>"""

        import sys, astropy
        hwin = HtmlWindow(self, wx.ID_ANY, size=(400,200))
        vers = {}
        from . import _version
        vers["webbpsf"] = _version.__version__
        try:
            import pysynphot
            vers['pysynphot'] = pysynphot.__version__
        except:
            vers['pysynphot'] = "Not Found"
        try:
            import pyfftw
            vers['pyfftw'] = "Present"
        except:
            vers['pyfftw'] = "Not Found"
        try:
            import fftw3
            vers['pyfftw3'] = "Present"
        except:
            vers['pyfftw3'] = "Not Found"
            
        vers["python"] = sys.version.split()[0]
        vers["wxpy"] = wx.VERSION_STRING
        vers['numpy'] = np.__version__
        vers['matplotlib'] = matplotlib.__version__
        #vers['atpy'] = atpy.__version__
        vers['astropy'] = astropy.__version__
        hwin.SetPage(aboutText % vers)
        btn = hwin.FindWindowById(wx.ID_OK)
        irep = hwin.GetInternalRepresentation()
        hwin.SetSize((irep.GetWidth()+25, irep.GetHeight()+10))
        self.SetClientSize(hwin.GetSize())
        self.CentreOnParent(wx.BOTH)
        self.SetFocus()


#-------------------------------------------------------------------------

def synplot(thing, waveunit='micron', label=None, **kwargs):
    """ Plot a single PySynPhot object (either SpectralElement or SourceSpectrum)
    versus wavelength, with nice axes labels.

    Really just a simple convenience function.
    """

    # convert to requested display unit.
    wave = thing.waveunits.Convert(thing.wave,waveunit)


    if label is None:
        label = thing.name


    if isinstance(thing, pysynphot.spectrum.SourceSpectrum):
        artist = plt.loglog(wave, thing.flux, label=label, **kwargs)
        plt.xlabel("Wavelength [%s]" % waveunit)
        if str(thing.fluxunits) == 'flam':
            plt.ylabel("Flux [%s]" % ' erg cm$^{-2}$ s$^{-1}$ Ang$^{-1}$' )
        else:
            plt.ylabel("Flux [%s]" % thing.fluxunits)
    elif isinstance(thing, pysynphot.spectrum.SpectralElement):
        artist = plt.plot(wave, thing.throughput,label=label, **kwargs)
        plt.xlabel("Wavelength [%s]" % waveunit)
        plt.ylabel("Throughput")
        plt.gca().set_ylim(0,1)
    else:
        _log.error( "Don't know how to plot that object...")
        artist = None
    return artist




def wxgui(fignum=1, showlog=True):
    import poppy
    # enable log message printout
    logging.basicConfig(level=logging.INFO,format='%(name)-10s: %(levelname)-8s %(message)s')


    # GUI does not play well with multiprocessing, so avoid that.
    if poppy.conf.use_multiprocessing:
        _log.error('Multiprocessing is not compatible with the GUI right now. Falling back to single-threaded.')
        poppy.conf.use_multiprocessing = False 

    # start the GUI
    app = wx.App()
    app.SetAppName("WebbPSF")
    gui = WebbPSF_GUI()
    gui.Show()

    if showlog: 
        # start it immediately below the main window
        gpos = gui.GetScreenPosition()
        gsize = gui.GetSize()

        logwin = LogFrame(pos=(gpos[0], gpos[1]+gsize[1]), size=(gsize[0], 200))
        logwin.Show()

        gui.log_window = logwin
    #bout = AboutBox()
    #about.ShowModal()
    #gui = WebbPSFOptionsDialog()
    #plt.figure(fignum)
    #plt.show(block=False)
    app.MainLoop()



if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO,format='%(name)-10s: %(levelname)-8s %(message)s')
    wxgui()






import os, sys
import ConfigParser

import logging
_log = logging.getLogger('webbpsf')

from . import conf


def restart_logging(verbose=True):
    """ Restart logging using the same settings as the last WebbPSF session, as stored in the configuration system. """

    level = str(conf.logging_level).upper()
    lognames = ['webbpsf', 'poppy']


    if level =='NONE':
        # disable logging
        lev = logging.CRITICAL  # we don't generate any CRITICAL flagged log items, so
                                # setting the level to this is effectively the same as ignoring
                                # all log events. FIXME there's likely a cleaner way to do this.
        if verbose: print "No log messages will be shown from WebbPSF."
    elif level in ['DEBUG', 'INFO','WARN','ERROR']:
        lev = logging.__dict__[level] # obtain one of the DEBUG, INFO, WARN, or ERROR constants
        if verbose: print "WebbPSF log messages of level {0} and above will be shown.".format(level)
    else:
        raise ValueError("Invalid logging level: "+level)
        return

    for name in lognames:
        logging.getLogger(name).setLevel(lev)

    # set up screen logging
    logging.basicConfig(level=logging.INFO,format='%(name)-10s: %(levelname)-8s %(message)s')
    if verbose: print("WebbPSF log outputs will be directed to the screen.")

    # set up file logging
    filename = conf.logging_filename
    if filename is None or filename.strip().lower() != 'none':
        hdlr = logging.FileHandler(filename)

        formatter = logging.Formatter('%(asctime)s %(name)-10s: %(levelname)-8s %(message)s')
        hdlr.setFormatter(formatter)

        for name in lognames:
            logging.getLogger(name).addHandler(hdlr)

        if verbose: print("WebbPSF log outputs will also be saved to file "+filename)


def setup_logging(level='INFO',  filename=None):
    """ Allows selection of logging detail and output locations (screen and/or file)

    This is a convenience wrapper to Python's built-in logging package, as used
    by webbpsf and poppy.  By default, this sets up log messages to be written
    to the screen, but the user can also request logging to a file. 

    The settings applied here are stored persistently between sessions using the
    astropy.config system. In many cases, you can just configure these once when you
    first install webbpsf and then logging will continue to operate as desired 
    without any additional intervention.


    For more advanced log handling, see the Python logging module's own documentation.
    
    Parameters
    -------------
    level : str
        name of log output to show. Defaults to 'INFO', set to 'DEBUG' for
        more extensive messages, or to "WARN" or "ERROR" for fewer. 
    filename : str, optional
        Filename to write the log output to. If not set, output will just 
        be displayed on screen. 


    Examples
    -----------

    >>> webbpsf.setup_logging(filename='webbpsflog.txt')

    This will save all log messages to 'webbpsflog.txt' in the current directory.
    If you later start another copy of webbpsf in a different directory, that session
    will also write to 'webbpsflog.txt' in *that* directory. Alternatively you can 
    specify a fully qualified absolute path to save all your logs to one specific file.


    >>> webbpsf.setup_logging(level='WARN')

    This will show only WARNING or ERROR messages on screen, and not save any logs to
    files at all (since the filename argument is None)

    
    """

    # implementation note: All this function actually does is apply the
    # defaults into the configuration system, then calls restart_logging to
    # do the actual work.

    level = str(level).upper()

    # The astropy config system will enforce the limited set of values for the logging_level parameter
    # by raising a TypeError on this next line if we feed in an invalid string.
    conf.logging_level = level


    if filename is None: filename='none' # must be a string to store into the config system
    conf.logging_filename = filename
    restart_logging(verbose=True)

MISSING_WEBBPSF_DATA_MESSAGE = """
 *********  ERROR  ******  ERROR  ******  ERROR  ******  ERROR  *************
 *                                                                          *
 *  WebbPSF requires several data files to operate.                         *
 *  These files could not be located automatically at this                  *
 *  time. Please download them to a location of your                        *
 *  choosing and either                                                     *
 *   - set the environment variable $WEBBPSF_PATH to point there, or        *
 *   - set the WEBBPSF_PATH variable in your webbpsf.cfg config file        *
 *     (probably in ~/.astropy/config/ or similar location)                 *
 *                                                                          *
 *  See http://pythonhosted.org/webbpsf/installation.html, under            *
 *  "Installing the Required Data Files", for more details.                 *
 *  WebbPSF will not be able to function properly until this has been done. *
 *                                                                          *
 ****************************************************************************
"""

def get_webbpsf_data_path():
    """ Get webbpsf data path

    Simply checking an environment variable is not always enough, since
    for packaging this code as a Mac .app bundle, environment variables are
    not available since .apps run outside the Terminal or X11 environments.

    Therefore, check first the environment variable WEBBPSF_PATH, and secondly
    check the configuration file in the user's home directory.
    """
    import os
    path_from_config = conf.WEBBPSF_PATH # read from astropy configuration
    if path_from_config == 'from_environment_variable':
        path = os.getenv('WEBBPSF_PATH')
        if path is None:
            raise EnvironmentError("Environment variable $WEBBPSF_PATH is not set!")
    else:
        path = path_from_config

    # at minimum, the path must be a valid directory
    if not os.path.isdir(path):
        raise IOError("WEBBPSF_PATH ({}) is not a valid directory path!".format(path))

    return path


def check_for_new_install(force=False):
    """ Check for a new installation, and if so
    print a hopefully helpful explanatory message.
    """

    from .version import version as __version__

    if os.getenv('WEBBPSF_SKIP_CHECK') is not None: return

    if conf.last_version_ran == '0.0' or force:

        from . import _save_config
        import astropy.config

        conf.last_version_ran = __version__
        _save_config()
        #conf.last_version_ran.save()
        #save_config('webbpsf') # save default values to text file

        print """
  ***************************************************
  *           WebbPSF Initialization & Setup         *
  ****************************************************

    This appears to be the first time you have used WebbPSF. 
    
    Just so you know, there is a mailing list for users of
    webbpsf to keep you informed about any announcements of
    updates or enhancements to this software. If you would like
    to subscribe, please email 
        majordomo@stsci.edu
    with the message 
        subscribe webbpsf-users

    
    WebbPSF has some options that can be set using a 
    configuration file. An example configuration file with
    default values has been created in 
            {0}/webbpsf.{1}.cfg
    (unless such a config file was already present there)
    You can examine that file and change settings if desired.
    See the WebbPSF documentation for more detail. """.format(astropy.config.get_config_dir(), __version__)

        # check for data dir?
        path_from_env_var = os.getenv('WEBBPSF_PATH') 

        if path_from_env_var is not None:
            print """

    WebbPSF's required data files appear to be 
    installed at a path given by $WEBBPSF_PATH :
    {0} """.format(path_from_env_var)
        else:
            # the following will automatically print an error message if
            # the path is unknown in the config file.
            path_from_config = conf.WEBBPSF_PATH

            if path_from_config != 'unknown':
                print """
    WebbPSF's required data files appear to be 
    installed at a path given in the config file:
    {0} """.format(path_from_config)

        print """

    This message will not be displayed again.
    Press [Enter] to continue
    """
        any_key = raw_input()

DIAGNOSTIC_REPORT = """
OS: {os}
Python version: {python}
numpy version: {numpy}
poppy version: {poppy}
webbpsf version: {webbpsf}

tkinter version: {tkinter}
wxpython version: {wxpython}

astropy version: {astropy}
pysynphot version: {pysyn}
pyFFTW version: {pyfftw}

Floating point type information for numpy.float:
{finfo_float}
Floating point type information for numpy.complex:
{finfo_complex}
"""

def system_diagnostic():
    """ return various helpful/informative information about the
    current system. For instance versions of python & available packages.

    Mostly undocumented function...
    """

    # There is probably a more clever way to do the following via introspection?

    import platform
    import os 
    import poppy
    import numpy
    from .version import version
    try:
        import ttk
        ttk_version = ttk.__version__
    except:
        ttk_version = 'not found'

    try:
        import wx
        wx_version = wx.__version__
    except:
        wx_version = 'not found'

    try:
        import pyfftw
        pyfftw_version = pyfftw.version
    except:
        pyfftw_version = 'not found'

    try:
        import pysynphot
        pysynphot_version = pysynphot.__version__
    except:
        pysynphot_version = 'not found'


    try:
        import astropy
        astropy_version = astropy.__version__
    except:
        astropy_version = 'not found'

    result = DIAGNOSTIC_REPORT.format(
        os=platform.platform(),
        numpy=numpy.__version__,
        python=sys.version.replace("\n", " "),
        poppy=poppy.__version__,
        webbpsf=version,
        tkinter=ttk_version,
        wxpython=wx_version,
        pyfftw=pyfftw_version,
        pysyn=pysynphot_version,
        astropy=astropy_version,
        finfo_float=numpy.finfo(numpy.float),
        finfo_complex=numpy.finfo(numpy.complex),
    )
    return result




""" jwxml: Various Python classes for parsing JWST-related information in XML files

* SUR: a segment update request file (mirror move command from the WAS to the MCS)
* Update: a single mirror update inside of a SUR
* SIAF: a SIAF file (Science Instrument Aperture File, listing the defined apertures for a given instrument)
* Aperture: a single aperture inside a SIAF


"""
import numpy as np
import matplotlib.pyplot as plt
try:
    from lxml import etree
    HAVE_LXML = True
except ImportError:
    import xml.etree.cElementTree as etree
    HAVE_LXML = False

import logging
import unittest
import os
_log = logging.getLogger('jwxml')


#---------------------------------------------------------------------------------
#   Mirror Move related classes

class Segment_Update(object):
    """ Class for representing one single mirror update (will be inside of groups in SURs)
    """
    def __init__(self, xmlnode):
        if xmlnode.attrib['type'] != 'pose': raise NotImplemented("Only Pose updates supported yet")

        self.id = int(xmlnode.attrib['id'])
        self.type = xmlnode.attrib['type']
        self.segment = xmlnode.attrib['seg_id'][0:2]
        self.absolute = xmlnode.attrib['absolute'] =='true'
        self.coord= xmlnode.attrib['coord'] #local or global
        self.stage_type= xmlnode.attrib['stage_type']  # recenter_fine, fine_only, none

        self.units = dict()
        self.moves = dict()
        for move in iterchildren(xmlnode):
            #print move.tag, move.text 
            self.moves[move.tag] =float(move.text)
            self.units[move.tag] = move.attrib['units']
            #X_TRANS, Y_TRANS, PISTON, X_TILT, Y_TILT, CLOCK
        #allowable units: 
		#units="id"
		#units="meters"
		#units="none"
		#units="radians"
		#units="sag"
		#units="steps"
		#
        # pose moves will only ever have meters/radians as units
    def __str__(self):
        return ("Update %d, %s, %s: "% (self.id, 'absolute' if self.absolute else 'relative', self.coord)) + str(self.moves)
    def shortstr(self):
        outstr = ("Update %d: %s, %s, %s {"% (self.id, self.segment, 'absolute' if self.absolute else 'relative', self.coord))

        outstr+= ", ".join([ coordname+"=%.3g" % self.moves[coordname] for coordname in ['PISTON','X_TRANS','Y_TRANS','CLOCK', 'X_TILT','Y_TILT']])
        #for coordname in ['PISTON','X_TRANS','Y_TRANS','CLOCK', 'X_TILT','Y_TILT']:
            #outstr+=coordname+"=%.3g" % self.moves[coordname]
        outstr+="}"
        return outstr

    @property
    def xmltext(self):
        """ The XML text representation of a given move """
        text= '        <UPDATE id="{0.id}" type="{0.type}" seg_id="{0.segment}" absolute="{absolute}" coord="{0.coord}" stage_type="{0.stage_type}">\n'.format( self, absolute = str(self.absolute).lower())
        for key in ['X_TRANS','Y_TRANS','PISTON','X_TILT', 'Y_TILT', 'CLOCK']:
            if key in self.moves.keys():
                text+='            <{key}  units="{unit}">{val:E}</{key}>\n'.format(key=key, unit=self.units[key], val=self.moves[key])
        text+= '        </UPDATE>\n'
        return text

    def toGlobal(self):
        """ Return moves cast to global coordinates """
        if self.coord =='global':
            return self.moves
        else:
            raise NotImplemented("Error")


    def toLocal(self):
        """ Return moves cast to local coordinates """
        if self.coord =='local':
            return self.moves
        else:
            raise NotImplemented("Error")
            # TO implement based on Ball's 'pmglobal_to_seg' in ./wfsc_core_algs/was_core_pmglobal_to_seg.pro
            # or the code in ./segment_control/mcs_hexapod_obj__define.pro


class SUR(object):
    """ Class for parsing/manipulating Segment Update Request files

    """
    def __init__(self, filename):
        """ Read a SUR from disk """
        self.filename=filename

        self._tree = etree.parse(filename)

        for tag in ['creator','date','time','version', 'operational']:
            self.__dict__[tag] = self._tree.getroot().attrib[tag]
        for element in self._tree.getroot().iter():
            if element.tag =='CONFIGURATION_NAME':  self.configuration_name = element.text
            if element.tag =='CORRECTION_ID':  self.correction_id = element.text

        self.groups = []
        for grp in self._tree.getroot().iter('GROUP'):
            myupdates = []
            for update in grp.iter('UPDATE'):
                #print update
                myupdates.append(Segment_Update(update))
            self.groups.append(myupdates)

    def __str__(self):
        outstr = "SUR %s\n" % self.filename #, type=%s, coords=%s\n" % (self.filename, 'absolute' if self.absolute else 'relative', self.coord)
        for igrp, grp in enumerate(self.groups):
            outstr+= "\tGroup %d\n" % (igrp+1)
            for update in grp:
                outstr+= "\t\t"+str(update)+"\n"
        return outstr

    @property
    def xmltext(self):
        """ The XML text representation of a given move """
        text = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<SEGMENT_UPDATE_REQUEST creator="?" date="{date}" time="{time}" version="0.0.1" operational="false" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="../../setup_files/schema/segment_update_request.xsd">
    <CONFIGURATION_NAME>{self.configuration_name}</CONFIGURATION_NAME>
    <CORRECTION_ID>{self.correction_id}</CORRECTION_ID>\n""".format(self=self, date='YYYY-MM-DD', time='HH:MM:SS')
    # FIXME add date and time keywords for real
        for igrp, grp in enumerate(self.groups):
            text+='    <GROUP id="{id}">\n'.format(id=igrp+1)
            for update in grp:
                text+=update.xmltext
            text+='    </GROUP>\n'
        text+= '</SEGMENT_UPDATE_REQUEST>'
        return text


    #@property
    #def name(self): return self._tree.getroot().attrib['name']


#---------------------------------------------------------------------------------
#  SIAF related classes

class Aperture(object):
    """ An Aperture, as parsed from the XML.
    All XML nodes are converted into object attributes. 

    See JWST-STScI-001550 for the reference on which this implementation was based. 

    4 Coordinate systems:
        * Detector:  pixels, in raw detector read out axes orientation ("Det")
        * Science:   pixels, in conventional DMS axes orientation ("Sci")
        * Ideal:     arcsecs relative to aperture reference location. ("Idl")
        * Telescope: arcsecs V2,V3 ("Tel")


    Example
    ========

    ap = some_siaf['desired_aperture_name']     # extract one aperture from a SIAF

    ap.Det2Tel(1024, 512)                       # convert pixel coordinates to sky Tel coords.
                                                # takes pixel coords, returns arcsec
    ap.Idl2Sci( 10, 3)                          # convert Idl coords to Sci pixels
                                                # takes arcsec, returns pixels

    # there exist functions for all of the possible {Tel,Idl,Sci,Det}2{Tel,Idl,Sci,Det} combinations.
    # you can also specify frames by string:
    ap.convert(1024, 512, frame_from='Det', frame_to='Tel')  # same as first example above

    ap.corners('Tel')                           # Get detector corners in Tel frame
    ap.center('Tel')                            # Get the reference point defined in the SIAF
                                                # this is typically the center of this region

    ap.plot('Idl', annotate=True, title=True)   # plot coordinates in Idl frame
    ap.plotDetectorChannels()                   # color in the readout channels


    """
    def __init__(self, xmlnode, instrument=None):

        self.instrument=instrument
        convfactors = {'RADIANS': 1, 'DEGREES': np.pi/180, 'ARCSECS': np.pi/180/60/60}

        for node in iterchildren(xmlnode): 
            tag = node.tag.replace('{http://www.stsci.edu/SIAF}','')
            if len(node.getchildren()) ==0:
                # if doens't have children, 
                try:
                    value = float(node.text) # do we care about ints vs floats?
                except: 
                    value=node.text
                self.__dict__[tag] = value
            else:
                # if does have children:
                if '{http://www.stsci.edu/SIAF}units' in [c.tag for c in node.getchildren()]:
                    # this will be an angle/units pair. units are either in arcsec or degrees. Convert to radians in any case for internal use.
                    unit  = node.find('{http://www.stsci.edu/SIAF}units').text
                    value =float( node.find('{http://www.stsci.edu/SIAF}value').text) * convfactors[unit]
                    self.__dict__[tag] = value
                elif '{http://www.stsci.edu/SIAF}elt' in [c.tag for c in node.getchildren()]:
                    #  an array of values which should go to an NDarray
                    elts = [float(c.text) for c in iterchildren(node, '{http://www.stsci.edu/SIAF}elt')]
                    self.__dict__[tag] = np.asarray(elts)

                else:
                    raise NotImplemented("Not sure how to parse that node.")

        # pack things into NDarrays for convenient access
        # first the vertices
        self.XIdlVert = np.asarray((self.XIdlVert1, self.XIdlVert2,self.XIdlVert3,self.XIdlVert4))
        self.YIdlVert = np.asarray((self.YIdlVert1, self.YIdlVert2,self.YIdlVert3,self.YIdlVert4))

        # then the transformation coefficients
        self.Sci2IdlDeg = int(self.Sci2IdlDeg)
        self.Sci2IdlCoeffs_X = np.zeros( (self.Sci2IdlDeg+1, self.Sci2IdlDeg+1))
        self.Sci2IdlCoeffs_Y = np.zeros( (self.Sci2IdlDeg+1, self.Sci2IdlDeg+1))
        self.Idl2SciCoeffs_X = np.zeros( (self.Sci2IdlDeg+1, self.Sci2IdlDeg+1))
        self.Idl2SciCoeffs_Y = np.zeros( (self.Sci2IdlDeg+1, self.Sci2IdlDeg+1))
        for i in range(1,self.Sci2IdlDeg+1):
            for j in range(0,i+1):
                #if self.AperName == 'FGS2_FULL_CNTR':
                    #print 'Sci2IdlX{0:1d}{1:1d}'.format(i,j), self.__dict__['Sci2IdlX{0:1d}{1:1d}'.format(i,j)]
                self.Sci2IdlCoeffs_X[i,j] = self.__dict__['Sci2IdlX{0:1d}{1:1d}'.format(i,j)]
                self.Sci2IdlCoeffs_Y[i,j] = self.__dict__['Sci2IdlY{0:1d}{1:1d}'.format(i,j)]
                self.Idl2SciCoeffs_X[i,j] = self.__dict__['Idl2SciX{0:1d}{1:1d}'.format(i,j)]
                self.Idl2SciCoeffs_Y[i,j] = self.__dict__['Idl2SciY{0:1d}{1:1d}'.format(i,j)]

    def __repr__(self):
        return "<jwxml.Aperture object AperName={0} >".format(self.AperName)


    #--- the actual fundamental transformation code follows in these next routines:
    def Det2Sci(self, XDet, YDet):
        """ Detector to Science, following Section 4.1 of JWST-STScI-001550"""
        XDet = np.asarray(XDet, dtype=float)
        YDet = np.asarray(YDet, dtype=float)
        ang = np.deg2rad(self.DetSciYAngle)
        XSci = self.XSciRef + self.DetSciParity* ((XDet - self.XDetRef)* np.cos(ang) + (YDet-self.YDetRef) * np.sin(ang))
        YSci = self.YSciRef -                     (XDet - self.XDetRef)* np.sin(ang) + (YDet-self.YDetRef) * np.cos(ang)
        return XSci, YSci

    def Sci2Det(self, XSci, YSci):
        """ Science to Detector, following Section 4.1 of JWST-STScI-001550"""
        XSci = np.asarray(XSci, dtype=float)
        YSci = np.asarray(YSci, dtype=float)
 
        ang = np.deg2rad(self.DetSciYAngle)
        XDet = self.XDetRef + self.DetSciParity * (XSci - self.XSciRef ) * np.cos(ang) - (YSci - self.YSciRef ) * np.sin(ang)
        YDet = self.YDetRef + self.DetSciParity * (XSci - self.XSciRef ) * np.sin(ang) + (YSci - self.YSciRef ) * np.cos(ang)
        return XDet, YDet

    def Sci2Idl(self, XSci, YSci):
        """ Convert Sci to Idl
        input in pixel, output in arcsec """
        dX = np.asarray(XSci, dtype=float) - self.XSciRef
        dY = np.asarray(YSci, dtype=float) - self.YSciRef
 
        degree = self.Sci2IdlDeg
        #CX = self.Sci2IdlCoefX
        #CY = self.Sci2IdlCoefY

        #XIdl = CX[0]*dX + CX[1]*dY + CX[2]*dX**2 + CX[3]*dX*dY + CX[4]*dY**2
        #YIdl = CY[0]*dY + CY[1]*dY + CY[2]*dY**2 + CY[3]*dY*dY + CY[4]*dY**2
        XIdl = np.zeros_like(np.asarray(XSci), dtype=float)
        YIdl = np.zeros_like(np.asarray(YSci), dtype=float)

        for i in range(1,degree+1):
            for j in range(0,i+1):
                XIdl += self.Sci2IdlCoeffs_X[i,j] * dX**(i-j) * dY**j
                YIdl += self.Sci2IdlCoeffs_Y[i,j] * dX**(i-j) * dY**j


        return XIdl, YIdl

    def Idl2Sci(self, XIdl, YIdl):
        """ Convert Idl to  Sci
        input in arcsec, output in pixels """
        XIdl = np.asarray(XIdl, dtype=float)
        YIdl = np.asarray(YIdl, dtype=float)
 
        degree = self.Sci2IdlDeg
        #dX = XIdl #Idl origin is by definition 0 
        #dY = YIdl #Idl origin is by definition 0

        XSci = np.zeros_like(np.asarray(XIdl), dtype=float)
        YSci = np.zeros_like(np.asarray(YIdl), dtype=float)

        for i in range(1,degree+1):
            for j in range(0,i+1):
                XSci += self.Idl2SciCoeffs_X[i,j] * XIdl**(i-j) * YIdl**j
                YSci += self.Idl2SciCoeffs_Y[i,j] * XIdl**(i-j) * YIdl**j



        #CX = self.Idl2SciCoefX
        #CY = self.Idl2SciCoefY

        #XSci = CX[0]*dX + CX[1]*dY + CX[2]*dX**2 + CX[3]*dX*dY + CX[4]*dY**2
        #YSci = CY[0]*dY + CY[1]*dY + CY[2]*dY**2 + CY[3]*dY*dY + CY[4]*dY**2
        return XSci + self.XSciRef, YSci + self.YSciRef
        #return XSci, YSci

    def Idl2Tel(self, XIdl, YIdl):
        """ Convert Idl to  Tel

        input in arcsec, output in arcsec

        WARNING
        --------
        This is an implementation of the planar approximation, which is adequate for most
        purposes but may not be for all. Error is about 1.7 mas at 10 arcminutes from the tangent
        point. See JWST-STScI-1550 for more details.
        """
        XIdl = np.asarray(XIdl, dtype=float)
        YIdl = np.asarray(YIdl, dtype=float)
 
        #print self.V2Ref, self.V3Ref
        #rad2arcsec = 1./(np.pi/180/60/60)

        #V2Ref and V3Ref are now in arcseconds in the XML file
        ang = np.deg2rad(self.V3IdlYAngle)
        V2 = self.V2Ref + self.VIdlParity * XIdl * np.cos(ang) + YIdl * np.sin(ang)
        V3 = self.V3Ref - self.VIdlParity * XIdl * np.sin(ang) + YIdl * np.cos(ang)
        return V2, V3

    def Tel2Idl(self,V2, V3):
        """ Convert Tel to Idl

        input in arcsec, output in arcsec

        This transformation involves going from global V2,V3 to local angles with respect to some
        reference point, and possibly rotating the axes and/or flipping the parity of the X axis.


        WARNING
        --------
        This is an implementation of the planar approximation, which is adequate for most
        purposes but may not be for all. Error is about 1.7 mas at 10 arcminutes from the tangent
        point. See JWST-STScI-1550 for more details.
        """
 
        #rad2arcsec = 1./(np.pi/180/60/60)
        dV2 = np.asarray(V2, dtype=float)-self.V2Ref
        dV3 = np.asarray(V3, dtype=float)-self.V3Ref
        ang = np.deg2rad(self.V3IdlYAng)

        XIdl = self.VIdlParity * (dV2 * np.cos(ang) - dV2 * np.sin(ang))
        YIdl =                    dV2 * np.sin(ang) + dV3 * np.cos(ang)
        return XIdl, YIdl

    #--- and now some compound transformations that are less fundamental. This just nests calls to the above.

    def Det2Idl(self, *args):
        return self.Sci2Idl(*self.Det2Sci(*args))
    def Det2Tel(self, *args):
        return self.Idl2Tel(*self.Sci2Idl(*self.Det2Sci(*args)))
    def Sci2Tel(self, *args):
        return self.Idl2Tel(*self.Sci2Idl(*args))
    def Idl2Det(self, *args):
        return self.Sci2Det(*self.Idl2Sci(*args))
    def Tel2Sci(self, *args):
        return self.Idl2Sci(*self.Tel2Idl(*args))
    def Tel2Det(self, *args):
        return self.Sci2Det(*self.Idl2Sci(*self.Tel2Idl(*args)))

    #--- now, functions other than direct coordinate transformations
    def convert(self, X, Y, frame_from=None, frame_to=None):
        """ Generic conversion routine, that calls one of the
        specific conversion routines based on the provided frame names as strings. """

        if frame_from is None: raise ValueError("You must specify a frame_from value : Tel, Idl, Sci, Det")
        if frame_to is None: raise ValueError("You must specify a frame_to value : Tel, Idl, Sci, Det")


        if frame_from == frame_to: return X, Y  # null transformation

        #frames = ['Det','Sci', 'Idl','Tel']
        function = eval('self.%s2%s' % (frame_from, frame_to))
        return function(X,Y)

    def corners(self, frame='Idl'):
        " Return coordinates of the aperture outline"
        return self.convert(self.XIdlVert, self.YIdlVert, 'Idl', frame)

    def center(self, frame='Tel'):
        """ Return the defining center point of the aperture"""
        return self.convert(self.V2Ref, self.V3Ref, 'Tel', frame)


    def plot(self, frame='Idl', label=True, ax=None, title=True, units='arcsec', annotate=False):
        """ Plot this one aperture 

        Parameters
        -----------
        frame : str
            Which coordinate system to plot in: 'Tel', 'Idl', 'Sci', 'Det'
        label : bool
            Add text label stating aperture name
        units : str
            one of 'arcsec', 'arcmin', 'deg'
        annotate : bool
            Add annotations for detector (0,0) pixels
        title : str
            If set, add a label to the plot indicating which frame was plotted.
 
        """


        if units is None:
            units='arcsec'

        # should we flip the X axis direction at the end of this function?
        need_to_flip_axis = False # only flip if we created the axis
        if ax is None:
            ax = plt.gca()
            ax.set_aspect('equal')
            if frame=='Idl' or frame=='Tel':
                need_to_flip_axis = True # *and* we're displaying some coordinates in angles relative to V2.
                ax.set_xlabel('V2 [{0}]'.format(units))
                ax.set_ylabel('V3 [{0}]'.format(units))

            elif frame=='Sci' or frame=='Det':
                ax.set_xlabel('X pixels [{0}]'.format(frame))
                ax.set_ylabel('Y pixels [{0}]'.format(frame))



        x, y = self.corners(frame=frame)

        if units.lower() == 'arcsec':
            scale=1
        elif units.lower() =='arcmin':
            scale=01./60
        elif units.lower() =='deg':
            scale=01./60/60
        else:
            raise ValueError("Unknown units: "+units)


        x2 = np.concatenate([x, [x[0]]]) # close the box
        y2 = np.concatenate([y, [y[0]]])
        ax.plot(x2*scale,y2*scale) # convert arcsec to arcmin


        if need_to_flip_axis:
            ax.set_xlim(ax.get_xlim()[::-1])

        if label:
            rotation = 30 if self.AperName.startswith('NRC') else 0 # partially mitigate overlapping NIRCam labels
            ax.text(x.mean()*scale, y.mean()*scale, self.AperName, 
                verticalalignment='center', horizontalalignment='center', rotation=rotation,
                color=ax.lines[-1].get_color())
        if title:
            ax.set_title("{0} frame".format(frame))
        if annotate:
            self.plotDetectorOrigin(frame=frame)

    def plotDetectorOrigin(self, frame='Idl', which='both'):
        """ Draw red and blue squares to indicate the raw detector 
        readout and science frame readout, respectively 

        Parameters
        -----------
        which : str
            Which detector origin to plot: 'both', 'Det', 'Sci'
        frame : str
            Which coordinate system to plot in: 'Tel', 'Idl', 'Sci', 'Det'
        """
 
        # raw detector frame
        if which.lower() == 'det' or which.lower()=='both':
            c1, c2 = self.convert( 0, 0, 'Det', frame)
            plt.plot(c1, c2, color='red', marker='s', markersize=9)

        # science frame
        if which.lower() == 'sci' or which.lower()=='both':
            c1, c2 = self.convert( 0, 0, 'Sci', frame)
            plt.plot(c1, c2, color='blue', marker='s')

    def plotDetectorChannels(self, frame='Idl', color='0.5', alpha=0.3, evenoddratio=0.5, **kwargs):
        """ Mark on the plot the various detector readout channels 

        These are depicted as alternating light/dark bars to show the
        regions read out by each of the output amps.

        Parameters
        ----------
        frame : str
            Optional if you have already called plot() to specify a 
            coordinate frame.
       """
 
        import matplotlib
        if self.instrument == 'MIRI': npixels = 1024 
        else: npixels = 2048
        ch = npixels/4
    
        ax = plt.gca()
        pts = ((0, 0), (ch,0), (ch,npixels), (0, npixels))
        for chan in range(4):
            plotpoints = np.zeros((4,2))
            for i,xy in enumerate(pts):
                plotpoints[i] = self.convert(xy[0]+chan*ch,xy[1],'Det',frame)
            rect = matplotlib.patches.Polygon(plotpoints, closed=True, 
                    alpha=(alpha if np.mod(chan,2) else alpha*evenoddratio), 
                    facecolor=color, edgecolor='none', lw=0)
            ax.add_patch(rect)


class SIAF(object):
    """ Science Instrument Aperture File 
    
    This is a class interface to SIAF information stored in an XML file. 
    It lets you read (only) the SIAF information, retrieve apertures, 
    plot them, and transform coordinates accordingly. 

    This class is basically just a container. See the Aperture class for
    the detailed implementation of the transformations. 

    Briefly, this class acts like a dict containing Aperture objects, accessible
    using their names defined in the SIAF

    Examples
    ---------

    fgs_siaf = SIAF('FGS')
    fgs_siaf.apernames                # returns a list of aperture names
    ap = fgs_siaf['FGS1_FULL_CNTR']   # returns an aperture object
    ap.plot(frame='Tel')              # plot one aperture
    fgs_siaf.plot()                   # plot all apertures in this file

    """
    def __init__(self, instr='NIRISS', basepath="./", filename=None):
            #basepath="/Users/mperrin/Dropbox/JWST/Optics Documents/SIAF/"
        """ Read a SIAF from disk 
        
        Parameters
        -----------
        instr : string
            one of 'NIRCam', 'NIRSpec', 'NIRISS', 'MIRI', 'FGS'; case sensitive.
        basepath : string
            Directory to look in for SIAF files
        filename : string, optional
            Alternative method to specify a specific SIAF XML file.
        """

        if instr not in ['NIRCam', 'NIRSpec', 'NIRISS', 'MIRI', 'FGS']:
            raise ValueError("Invalid instrument name: {0}. Note that this is case sensitive.".format(instr))

        self.instrument=instr

        if filename is None:
            self.filename=os.path.join(basepath, instr+'_SIAF.xml')
        else:
            self.filename = filename

        self.apertures = {}

        self._tree = etree.parse(self.filename)

        self._last_plot_frame=None



        #for entry in self._tree.getroot().iter('{http://www.stsci.edu/SIAF}SiafEntry'):
        for entry in self._tree.getroot().iter('SiafEntry'):
            aperture = Aperture(entry, instrument=self.instrument)
            self.apertures[aperture.AperName] = aperture

    def __getitem__(self, key):
        return self.apertures[key]

    def __len__(self):
        return len(self.apertures)

    @property
    def apernames(self):
        """ List of aperture names defined in this SIAF"""
        return self.apertures.keys()

 
    def _getFullApertures(self):
        """ Return whichever subset of apertures correspond to the entire detectors. This is a
        helper function for the various plotting routines following"""
        fullaps = []
        if self.instrument =='NIRCam':
            for letter in ['A', 'B']:
                for number in range(1,6):
                    fullaps.append(self.apertures['NRC{letter}{number}_FULL_CNTR'.format(letter=letter, number=number)])
        elif self.instrument =='NIRSpec':
            fullaps.append( self.apertures['NRS1_FULL_CNTR'])
            fullaps.append( self.apertures['NRS2_FULL_CNTR'])
        elif self.instrument =='NIRISS':
            fullaps.append( self.apertures['NIS_FULL_CNTR'])
        elif self.instrument =='MIRI':
            fullaps.append( self.apertures['MIRIM_FULL_CNTR_OSS'])
        elif self.instrument =='FGS':
            fullaps.append( self.apertures['FGS1_FULL_CNTR'])
            fullaps.append( self.apertures['FGS2_FULL_CNTR'])
        return fullaps


    def plot(self, frame='Tel', names=None, label=True, units=None, clear=True, annotate=False, 
        subarrays=True):
        """ Plot all apertures in this SIAF

        Parameters
        -----------
        names : list of strings
            A subset of aperture names, if you wish to plot only a subset
        subarrays : bool
            Plot all the minor subarrays if True, else just plot the "main" apertures
        label : bool
            Add text labels stating aperture names
        units : str
            one of 'arcsec', 'arcmin', 'deg'
        clear : bool
            Clear plot before plotting (set to false to overplot)
        annotate : bool
            Add annotations for detector (0,0) pixels
        frame : str
            Which coordinate system to plot in: 'Tel', 'Idl', 'Sci', 'Det'
        """
        if clear: plt.clf()
        ax = plt.subplot(111)
        ax.set_aspect('equal')

        # which list of apertures to iterate over?
        if subarrays:
            iterable = self.apertures.itervalues
        else: 
            iterable = self._getFullApertures

        for ap in iterable():
            if names is not None:
                if ap.AperName not in names: continue

            ap.plot(frame=frame, label=label, ax=ax, units=None)
            if annotate: 
                ap.plotDetectorOrigin(frame=frame)
        ax.set_xlabel('V2 [arcsec]')
        ax.set_ylabel('V3 [arcsec]')



        if frame =='Tel' or frame=='Idl':
            # enforce V2 increasing toward the left
            ax.autoscale_view(True,True,True)
            xlim = ax.get_xlim()
            if xlim[1] > xlim[0]: ax.set_xlim(xlim[::-1])
            ax.set_autoscalex_on(True)

        self._last_plot_frame = frame

    def plotDetectorOrigin(self, which='both', frame=None):
        """ Mark on the plot the detector's origin in Det and Sci coordinates 
        
        Parameters
        -----------
        which : str
            Which detector origin to plot: 'both', 'Det', 'Sci'
        frame : str
            Which coordinate system to plot in: 'Tel', 'Idl', 'Sci', 'Det'
            Optional if you have already called plot() to specify a 
            coordinate frame.
 
        """
        if frame is None: frame = self._last_plot_frame
        for ap in self._getFullApertures():
            ap.plotDetectorOrigin(frame=frame, which=which)


    def plotDetectorChannels(self, frame=None):
        """ Mark on the plot the various detector readout channels 
       
        These are depicted as alternating light/dark bars to show the
        regions read out by each of the output amps.

        Parameters
        ----------
        frame : str
            Which coordinate system to plot in: 'Tel', 'Idl', 'Sci', 'Det'
            Optional if you have already called plot() to specify a 
            coordinate frame.
  
        """
        if frame is None: frame = self._last_plot_frame
        for ap in self._getFullApertures():
            ap.plotDetectorChannels(frame=frame)


def plotAllSIAFs(subarrays = True, showorigin=True, showchannels=True, **kwargs):
    """ Plot All instrument """

    for instr in ['NIRCam','NIRISS','NIRSpec','FGS','MIRI']:
        aps =SIAF(instr, **kwargs)
        print "{0} has {1} apertures".format(aps.instrument, len(aps))

        aps.plot(clear=False, subarrays=subarrays)
        if showorigin: aps.plotDetectorOrigin()
        if showchannels: aps.plotDetectorChannels()


class Test_SIAF(unittest.TestCase):

    def assertAlmostEqualTwo(self, tuple1, tuple2):
        self.assertAlmostEqual(tuple1[0], tuple2[0], places=1)
        self.assertAlmostEqual(tuple1[1], tuple2[1], places=1)


    def _test_up(self):
        siaf = SIAF("JwstSiaf-2010-10-05.xml")
        startx = 1023
        starty = 1024

        nca = siaf['NIRCAM A']

        self.assertAlmostEqualTwo( nca.Det2Sci(startx,starty), (1020.,1020.))
        print "Det2Sci OK"

        self.assertAlmostEqualTwo( nca.Det2Idl(startx,starty), (0.0, 0.0))
        print "Det2Idl OK"
        self.assertAlmostEqualTwo( nca.Det2Tel(startx,starty), (87.50, -497.10))
        print "Det2Tel OK"

    def _test_down(self):
        siaf = SIAF("JwstSiaf-2010-10-05.xml")
        startV2 = 87.50
        startV3 = -497.10
        nca = siaf['NIRCAM A']

        self.assertAlmostEqualTwo( nca.Sci2Det(1020., 1020), (1023.,1024.))
        print "Sci2Det OK"

        self.assertAlmostEqualTwo( nca.Tel2Idl(startV2, startV3), (0.0, 0.0))
        print "Tel2Idl OK"
        self.assertAlmostEqualTwo( nca.Tel2Sci(startV2, startV3), (1020., 1020.))
        print "Tel2Sci OK"
        self.assertAlmostEqualTwo( nca.Tel2Det(startV2, startV3), (1023.,1024.))
        print "Tel2Det OK"

    def test_inverses(self):
        siaf = SIAF("JwstSiaf-2010-10-05.xml")
        nca = siaf['NIRCAM A']

        self.assertAlmostEqualTwo( nca.Det2Sci(*nca.Sci2Det(1020., 1020)), (1020., 1020) )
        self.assertAlmostEqualTwo( nca.Sci2Det(*nca.Det2Sci(1020., 1020)), (1020., 1020) )
        print "Det <-> Sci OK"

        self.assertAlmostEqualTwo( nca.Tel2Idl(*nca.Idl2Tel(10., 10)), (10., 10) )
        self.assertAlmostEqualTwo( nca.Idl2Tel(*nca.Tel2Idl(10., 10)), (10., 10) )
        print "Tel <-> Idl OK"

        self.assertAlmostEqualTwo( nca.Tel2Sci(*nca.Sci2Tel(10., 10)), (10., 10) )
        self.assertAlmostEqualTwo( nca.Sci2Tel(*nca.Tel2Sci(10., 10)), (10., 10) )
        print "Tel <-> Sci OK"


# The ElementTree implementation in xml.etree does not support
# Element.iterchildren, so provide this wrapper instead
# This wrapper does not currently provide full support for all the arguments as
# lxml's iterchildren
def iterchildren(element, tag=None):
    if HAVE_LXML:
        return element.iterchildren(tag)
    else:

        if tag is None:
            return iter(element)

        def _iterchildren():
            for child in element:
               if child.tag == tag:
                   yield child

        return _iterchildren()



if __name__== "__main__":
    logging.basicConfig(level=logging.DEBUG,format='%(name)-10s: %(levelname)-8s %(message)s')
    #unittest.main()
    #sur = SUR('/itar/jwst/wss/MMS_Delivery/09_MMS_Source_Code/wfsc_mcs~1.1.2/wfsc_mcs/fqt/tc7/sur_ok_rel_gl.xml')

    s = SIAF()



